from __future__ import print_function, absolute_import, division

def clapp():
    # Load system modules.
    from argparse import ArgumentParser
    import datetime
    import os
    try:
        import readline
    except ImportError:
        readline = None
    import signal
    import sys
    import warnings

    # Load custom modules
    from parmed.exceptions import ParmedError

    from parmed.tools.logos import Logo
    from parmed.tools.exceptions import SeriousParmWarning, InterpreterError
    from parmed.tools.parmed_cmd import ParmedCmd
    from parmed.tools.actions import Action
    from parmed.tools.parmlist import ParmList
    from parmed import __version__

    # Set up new excepthook to clean up fatal exception printouts
    def interrupted(*args, **kwargs):
        """ Handle interruptions gracefully """
        sys.stdout.write('Interrupted\n')
        sys.exit(1)

    signal.signal(signal.SIGINT, interrupted)

    # Define our own custom warning printer
    def _print_warnings(message, category, filename, lineno, file=None, line=None):
        """ Override the default showwarning method """
        if file is None: file = sys.stderr
        try:
            file.write('%s: %s\n' % (category.__name__, message))
        except IOError:
            pass

    warnings.showwarning = _print_warnings

    # Set up parser
    parser = ArgumentParser()
    parser.add_argument('-v', '--version', action='version',
             version='%%(prog)s: Version %s' % __version__)
    group = parser.add_argument_group('Input Files')
    group.add_argument('-i', '--input', dest='script', default=[],
             metavar='FILE', help='''Script with ParmEd commands to execute. Default
             reads from stdin. Can be specified multiple times to process multiple
             input files.''', action='append')
    group.add_argument('-p', '--parm', dest='prmtop', default=[],
             metavar='<prmtop>', action='append', help='''List of topology files to
             load into ParmEd. Can be specified multiple times to process multiple
             topologies.''')
    group.add_argument('-c', '--inpcrd', dest='inpcrd', default=[],
             metavar='<inpcrd>', action='append', help='''List of inpcrd files to
             load into ParmEd. They are paired with the topology files in the same
             order that each set of files is specified on the command-line.''')
    group = parser.add_argument_group('Output Files')
    group.add_argument('-O', '--overwrite', dest='overwrite', default=False,
             help='Allow ParmEd to overwrite existing files.', action='store_true')
    group.add_argument('-l', '--logfile', dest='logfile', metavar='FILE',
             default='parmed.log', help='''Log file with every command executed
             during an interactive ParmEd session. Default is parmed.log''')
    group = parser.add_argument_group('Interpreter Options', '''These options affect
             how the ParmEd interpreter behaves in certain cases.''')
    group.add_argument('--prompt', dest='prompt', default='>', metavar='PROMPT',
             help='String to use as a command prompt.')
    group.add_argument('-n', '--no-splash', dest='printlogo', action='store_false',
             help='Prevent printing the greeting logo.', default=True)
    group.add_argument('-e', '--enable-interpreter', dest='interpreter',
             action='store_true', default=False,
             help='''Allow arbitrary single Python commands or blocks of Python code
             to be run. By default Python commands will not be run as a safeguard
             for your system. Make sure you trust the source of the ParmEd command
             before turning this option on.''')
    group = parser.add_argument_group('Error Handling', '''These options control how
             ParmEd handles various errors and warnings that appear occur during the
             course of Action execution''')
    group.add_argument('-s', '--strict', dest='strict', action='store_true',
             default=True, help='''Prevent scripts from running past unrecognized
             input and actions that end with an error. In interactive mode, actions
             with unrecognized inputs and failed actions prevent any changes from
             being made to the topology, but does not quit the interpreter. This is
             the default behavior.''')
    group.add_argument('-r', '--relaxed', dest='strict', action='store_false',
             help='''Scripts ignore unrecognized input and simply skip over failed
             actions, executing the rest of the script. Unrecognized input in the
             interactive interpreter emits a non-fatal warning.''')
    parser.add_argument('prmtop_cl', nargs='?', metavar='<prmtop>', default=None,
             help='Topology file to analyze.')
    parser.add_argument('script_cl', nargs='?', metavar='<script>', default=None,
             help='File with a series of ParmEd commands to execute.')

    opt = parser.parse_args()

    # If the user specified a prmtop and script in the 'old' way, append them to the
    # list constructed via the --parm and -i flags -- they come at the end
    if opt.script_cl is not None:
        opt.script.append(opt.script_cl)

    if opt.prmtop_cl is not None:
        opt.prmtop.append(opt.prmtop_cl)

    # Load the splash screen
    if opt.printlogo:
        splash = Logo()
        print(splash)

    # Set our warning filter
    if not opt.strict:
        warnings.filterwarnings('always', category=SeriousParmWarning)

    # Set our overwrite preferences
    Action.overwrite = opt.overwrite

    amber_prmtop = ParmList()
    for i, parm in enumerate(opt.prmtop):
        if i < len(opt.inpcrd):
            amber_prmtop.add_parm(parm)
            amber_prmtop.parm.load_rst7(opt.inpcrd[i])
            print('Loaded Amber topology file %s with coordinates from %s\n' %
                  (parm, opt.inpcrd[i]))
        else:
            amber_prmtop.add_parm(parm)
            print('Loaded Amber topology file %s' % parm)

    if len(opt.script) > 0:
        # Read from the list of scripts
        print(opt.script)
        # Make sure that all scripts exist, quitting if we are strict and not all
        # scripts exist. Don't do anything until we know that all scripts exist.
        for script in opt.script:
            if not os.path.exists(script):
                warnings.warn('Script file %s cannot be found.' % script,
                              SeriousParmWarning)

        # We have already pre-screened the scripts.
        for script in opt.script:
            if not os.path.exists(script): continue
            print('Reading actions from %s\n' % script)
            parmed_commands = ParmedCmd(amber_prmtop, stdin=open(script, 'r'))
            parmed_commands.use_rawinput = 0
            parmed_commands.interpreter = opt.interpreter
            parmed_commands.prompt = ''
            # Loop through all of the commands
            try:
                parmed_commands.cmdloop()
            except InterpreterError as err:
                sys.exit('%s: %s' % (type(err).__name__, err))
            except ParmedError:
                # This has already been caught and printed. If it was re-raised,
                # then that means we wanted to exit
                sys.exit(1)

    else:
        close_log_file = False
        parmed_commands = ParmedCmd(amber_prmtop)
        parmed_commands.intro = "Reading input from STDIN..."
        parmed_commands.interpreter = opt.interpreter
        parmed_commands.prompt = opt.prompt.strip() + ' '
        # Set the log file and logging, but only if interactive
        if os.isatty(sys.stdin.fileno()):
            if os.path.exists(opt.logfile) and readline is not None:
                # Load the logfile as a history, and get rid of any timestamps
                # from the history
                try:
                    f = open(opt.logfile, 'r')
                except IOError:
                    pass # Do not have permission or something
                else:
                    for line in f:
                        if line.startswith('# Log started on'): continue
                        readline.add_history(line.rstrip())
                    f.close()
            try:
                logfile = open(opt.logfile, 'a')
            except IOError:
                print("Could not open logfile. Skip logging", file=sys.stderr)
                close_log_file = False
            else:
                now = datetime.datetime.now()
                logfile.write('# Log started on %02d/%02d/%d [mm/dd/yyyy] at '
                              '%02d:%02d:%02d\n' % (now.month, now.day, now.year,
                                                now.hour, now.minute, now.second))
                if len(sys.argv) > 1:
                    logfile.write('# Command line arguments: %s\n' %
                            ' '.join(sys.argv[1:]))
                if len(amber_prmtop) > 0:
                    logfile.write('# Loaded topologies: %s\n' % ', '.join(
                            name for name in amber_prmtop._parm_names))
                parmed_commands.setlog(logfile)
                close_log_file = True
        # Loop through all of the commands
        try:
            try:
                parmed_commands.cmdloop()
            except InterpreterError as err:
                sys.exit('%s: %s' % (type(err).__name__, err))
            except ParmedError:
                # This has already been caught and printed. If it was re-raised,
                # then that means we wanted to exit
                sys.exit(1)
        finally:
            if close_log_file:
                logfile.close()

    print('Done!')

def guiapp():
    from parmed import load_file
    from parmed.utils.six.moves import tkinter as tk
    from parmed.utils.six.moves.tkinter_messagebox import showerror
    from optparse import OptionParser
    from os.path import split
    from parmed.tools.exceptions import ParmError
    from parmed.tools.gui.guitools import ParmedApp
    from parmed.tools.gui.guifiletools import file_chooser
    from parmed.tools.logos import Logo
    from parmed.tools.actions import Action
    from parmed.tools.parmlist import ParmList
    import sys

    debug = False

    def excepthook(exc, msg, tb):
        """ Default exception handler """
        import traceback
        if debug:
            traceback.print_tb(tb)
        showerror('Fatal Error', '%s: %s' % (exc.__name__, msg))
        sys.exit(1)

    # Launch the root window
    root = tk.Tk()
    root.resizable(True, True)

    # Replace the default excepthook with mine
    sys.excepthook = excepthook

    # See if we were provided a topology file on the command-line
    parser = OptionParser(usage = '%prog [<prmtop>]')
    parser.add_option('-d', '--debug', dest='debug', default=False,
                      action='store_true', help='Show detailed tracebacks ' +
                      'when an error is detected.')
    opt, args = parser.parse_args()

    debug = opt.debug

    # If the user provided a CL argument, that is the prmtop_name. Otherwise,
    # open up a file choosing dialog box to get the input from the user
    if len(args) == 0:
        prmtop_name = file_chooser('Topology')
    elif len(args) == 1:
        prmtop_name = args[0]
    else:
        sys.stderr.write('Unknown command-line options. Ignoring\n')
        prmtop_name = file_chooser('Topology')

    # If we chose no prmtop file, throw an error
    if not prmtop_name:
        raise ParmError('No prmtop chosen!')

    # Load the amber prmtop and check for errors
    amber_prmtop = ParmList()
    parm = load_file(prmtop_name)
    amber_prmtop.add_parm(parm)

    # Make this overwritable -- the all of the file save boxes will ask the user
    # for verification before saving over an existing file. There's no need for
    # the AmberParm security layer.
    Action.overwrite = True

    fname = split(prmtop_name)[1]
    root.title('xParmED: Editing/viewing [%s] Choose an operation' % fname)

    # Now build the action list on root
    app = ParmedApp(root, amber_prmtop)
    app.pack(fill=tk.BOTH, expand=1)
    root.mainloop()

    print('Thank you for using xParmEd!\n%s' % Logo())



"""
This sets up the command interpreter for textual ParmEd (parmed).
"""

# Load some system modules that may be useful for various users in shell mode
import cmd
import os
import sys
import traceback
from glob import glob

from ..exceptions import ParmedError, ParmedWarning
from ..utils.six import iteritems
from ..utils.six.moves import range  # pylint: disable=W0622,E0401
from .actions import COMMANDMAP, Usages
from .argumentlist import ArgumentList
from .exceptions import InterpreterError

try:
    import readline
except ImportError:
    readline = None

_COMMANDLOGS = []

def log_commands(func):
    """ Decorator to write the line to a file """
    global _COMMANDLOGS
    def new_func(self, line, *args, **kwargs):
        if self._logfile is not None and line != 'EOF':
            self._logfile.write('%s\n' % line)
            if readline is None:
                _COMMANDLOGS.append(line)
            try:
                self._logfile.flush()
            except AttributeError:
                pass
        return func(self, line, *args, **kwargs)

    return new_func

class ParmedCmd(cmd.Cmd):
    """
    ParmEd command interpreter. The docstrings for each do_* command are simply
    copied from the docstring for that Action's docstring
    """

    prompt = "> "
    _populated = False

    def __init__(self, parm, stdin=None, stdout=None):
        """ Load a topology file into the interpreter """
        self.parm = parm
        cmd.Cmd.__init__(self, stdin=stdin, stdout=stdout)
        self.continued = ''
        self._logfile = None
        try:
            self._exit_on_fatal = not os.isatty(self.stdin.fileno())
        except AttributeError:
            self._exit_on_fatal = False
        if not self._populated: self.populate_actions()

    def setlog(self, f):
        """ Open up a log file to start logging the commands. """
        if f is None:
            return
        if hasattr(f, 'write'):
            self._logfile = f
        else:
            self._logfile = open(f, 'w')

    def emptyline(self):
        """ Override emptyline so that empty lines are simply ignored """
        pass

    @log_commands
    def precmd(self, line):
        """ Override this to strip out comments, but preserve #s in quotes """
        in_quotes = False
        quote_type = ''
        idx = -1
        for i, char in enumerate(line):
            if in_quotes:
                if char == quote_type:
                    in_quotes = False
                    continue
                continue
            if char == '"':
                in_quotes = True
                quote_type = '"'
                continue
            if char == "'":
                in_quotes = True
                quote_type = "'"
                continue
            if char == '#':
                idx = i
                break

        if idx < 0:
            if line.endswith('\\'):
                self.continued += line[:-1] + ' '
                return ''
            line = self.continued + line
            self.continued = ''
            return line
        else:
            line = self.continued + line[:idx]
            self.continued = ''
            return line

    def parseline(self, line):
        """
        Override parseline so that it will set args as ArgumentList
        """
        line = line.strip()
        if not line:
            return None, None, line
        elif line[0] == '?':
            line = 'help ' + line[1:]
        elif line[-1] == '?' and len(line.split()) == 1:
            line = 'help ' + line[:-1]
        elif line[0] == '!':
            if hasattr(self, 'do_shell'):
                line = 'shell ' + line[1:]
            else:
                return None, None, line
        i, n = 0, len(line)
        while i < n and line[i] in self.identchars:
            i = i+1
        cmd, arg = line[:i], ArgumentList(line[i:].strip())
        if len(line) == 4 and line == 'help':
            arg = ''
        if line[:5] == 'help ':
            arg = line[i:].strip()
        return cmd, arg, line

    def do_parmout(self, line):
        # Store this action for later use. This action is unique
        try:
            self.parmout = COMMANDMAP['parmout'](self.parm, line)
        except (ParmedError, ParmedWarning) as err:
            self.stdout.write('Action parmout failed.\n\t')
            self.stdout.write('%s: %s\n' % (type(err).__name__, err))
            if self._exit_on_fatal:
                raise
        except Exception as err:
            if self._exit_on_fatal:
                raise
            self.stdout.write('Unexpected failure:\n%s: %s\n\n'
                              'Traceback is\n\n' % (type(err).__name__, err))
            traceback.print_tb(sys.exc_info()[2], file=self.stdout)

    def _normaldo(self, ActionClass, line):
        """ The standard action command does this stuff """
        actionname = ActionClass.__name__
        try:
            action = ActionClass(self.parm, line)
            if action.valid:
                self.stdout.write('%s\n' % action)
                action.execute()
        except (ParmedError, ParmedWarning) as err:
            self.stdout.write('Action %s failed\n\t' % actionname)
            self.stdout.write('%s: %s\n' % (type(err).__name__, err))
            if self._exit_on_fatal:
                raise
        except Exception as err:
            if self._exit_on_fatal:
                raise
            self.stdout.write('Unexpected failure:\n%s: %s\n\n'
                              'Traceback is\n\n' % (type(err).__name__, err))
            traceback.print_tb(sys.exc_info()[2], file=self.stdout)

    @classmethod
    def populate_actions(cls):
        """
        This will create all of the do_Command methods to trigger command
        auto-complete. This eliminates the need to modify the ParmedCmd class
        when a new command is added
        """
        for _cmd, cmdclass in iteritems(COMMANDMAP):
            if _cmd in ('source', 'go', 'EOF', 'quit', 'help', 'parmout'):
                continue
            cmdname = cmdclass.__name__
            exec('cls.do_%s = lambda self, line: '
                 'self._normaldo(COMMANDMAP["%s"], line)' % (cmdname, _cmd))
        cls._populated = True

    def do_source(self, line):
        action = COMMANDMAP['source'](self.parm, line)
        if not action.valid: return
        self.stdout.write('%s\n' % action)
        _cmd = ParmedCmd(self.parm, stdin=open(action.filename, 'r'),
                         stdout=self.stdout)
        _cmd.prompt = ''
        _cmd.interpreter = self.interpreter
        _cmd.use_rawinput = 0
        # If we exit on error, call cmdloop without try protection. Otherwise,
        # we need to protect cmdloop to catch any ParmedError's that are passed
        # through to avoid exiting the top-level interpreter
        if self._exit_on_fatal:
            _cmd.cmdloop()
        else:
            # If we are not exiting on error, just ignore these parm errors
            # since they've already been printed. Now just wait for the next
            # command from the user.
            try:
                _cmd.cmdloop()
            except:
                pass

    def do_go(self, line):
        """
        Stops reading commands and executes any 'parmout' command that had
        been issued
        """
        if hasattr(self, 'parmout'):
            self.stdout.write('%s\n' % self.parmout)
            try:
                self.parmout.execute()
            except ParmedError as err:
                self.stdout.write('%s: %s\n' % (type(err).__name__, err))
                if self._exit_on_fatal:
                    raise
            except Exception as err:
                if self._exit_on_fatal:
                    raise
                self.stdout.write('Unexpected failure:\n%s: %s\n\nTraceback is'
                                  '\n\n' % (type(err).__name__, err))
                traceback.print_tb(sys.exc_info()[2], file=self.stdout)

        return True

    # EOF is treated the same as "go"
    do_EOF = do_go

    def do_quit(self, line):
        """
        Stops reading commands and quits WITHOUT running the last 'parmout' 
        command that had been issued
        """
        return True

    def do_history(self, line):
        """
        Print the command history
        """
        global _COMMANDLOGS
        if readline is None:
            for line in _COMMANDLOGS:
                self.stdout.write('%s\n' % line)
        else:
            for i in range(readline.get_current_history_length()):
                self.stdout.write('%s\n' % readline.get_history_item(i+1))

    def default(self, line):
        words = line.split()
        mycmd = words[0].lower()
        if mycmd not in COMMANDMAP:
            self.stdout.write("%s command not recognized\n" % words[0])
        else:
            args = ArgumentList(words[1:])
            self._normaldo(COMMANDMAP[mycmd], args)

    def do_shell(self, line):
        """ Support the limited interpreter """
        if not self.interpreter:
            raise InterpreterError("Interpreter not enabled! Use '-e' "
                                   "to enable")
        line = str(line)
        if line.strip() == '!':
            self._python_shell()
        else:
            globals_ = dict(amber_prmtop=self.parm)
            globals_.update(globals())
            try:
                exec(line.strip(), globals_)
            except Exception as err:
                self.stdout.write("%s: %s\n" % (type(err).__name__, err))

    def completedefault(self, text, line, begidx, endidx):
        partial = line[:endidx]
        idx = max(partial.rfind(' '), partial.rfind('\t')) + 1
        beg_token = partial[idx:begidx]
        return [s.replace(beg_token, '', 1) for s in glob(partial[idx:] + '*')]

    def _python_shell(self):
        """ Drop into a limited interpreter and read until we see !! """
        from parmed.tools import actions
        python_interpreter = PythonCmd(stdin=self.stdin, stdout=self.stdout)
        python_interpreter.use_rawinput = self.use_rawinput
        python_interpreter.setlog(self._logfile)
        if not self.prompt: python_interpreter.prompt = ''
        python_interpreter.cmdloop()
        try:
            globals_ = dict(amber_prmtop=self.parm, actions=actions)
            globals_.update(globals())
            exec(python_interpreter.command_string, globals_)
        except Exception as err:
            self.stdout.write("%s: %s\n" % (type(err).__name__, err))
      
    def do_help(self, arg):
        " Modify the original do_help to pull docstrings from actions "
        if arg:
            # XXX check arg syntax
            try:
                func = getattr(self, 'help_' + arg)
            except (AttributeError, TypeError):
                try:
                    doc=getattr(self, 'do_' + arg).__doc__
                    if doc:
                        self.stdout.write("%s\n" % str(doc))
                        return
                except (AttributeError, TypeError):
                    pass
                try:
                    _action = COMMANDMAP[arg.lower()]
                    if _action.needs_parm:
                        doc = '%s [parm <idx>|<name>]\n%s'
                    else:
                        doc = '%s\n%s'
                    doc = doc % (_fmt_usage(Usages[arg.lower()]),
                                 _action.__doc__)
                    if doc:
                        self.stdout.write("%s\n"%str(doc))
                        return
                except (AttributeError, KeyError):
                    pass
                self.stdout.write("%s\n" % str(self.nohelp % (arg,)))
                return
                func()
        else:
            names = self.get_names()
            cmds_doc = []
            cmds_undoc = []
            help = {}
            for name in names:
                if name[:5] == 'help_':
                    help[name[5:]]=1
            names.sort()
            # There can be duplicates if routines overridden
            prevname = ''
            for name in names:
                if name[:3] == 'do_':
                    if name == prevname:
                        continue
                    prevname = name
                    cmd=name[3:]
                    if cmd in help:
                        cmds_doc.append(cmd)
                        del help[cmd]
                    elif getattr(self, name).__doc__:
                        cmds_doc.append(cmd)
                    elif cmd.lower() in COMMANDMAP and cmd.lower() in Usages:
                        cmds_doc.append(cmd)
                    else:
                        cmds_undoc.append(cmd)
            self.stdout.write("%s\n"%str(self.doc_leader))
            self.print_topics(self.doc_header,   cmds_doc,   15,80)
            self.print_topics(self.misc_header,  help.keys(),15,80)
            self.print_topics(self.undoc_header, cmds_undoc, 15,80)

class PythonCmd(cmd.Cmd):
    """ Command interpreter for limited Python interpreter """
    prompt = "py >>> "

    def __init__(self, stdin=None, stdout=None):
        cmd.Cmd.__init__(self, stdin=stdin, stdout=stdout)
        self._logfile = None
        self.command_string = ""

    def setlog(self, f):
        """ Open up a log file to start logging the commands """
        if f is None: return
        if hasattr(f, 'write'):
            self._logfile = f
        else:
            self._logfile = open(f, 'w')

    precmd = log_commands(cmd.Cmd.precmd)

    def emptyline(self):
        """ Ignore all empty lines """
        self.command_string += "\n"

    def do_shell(self, line):
        """ Break out of the shell """
        return True

    def do_EOF(self, line):
        """ Break out of the shell """
        raise InterpreterError("EOF hit while in interpreter!")

    def parseline(self, line):
        """ Override parseline """
        if not line.strip():
            return False, False, False
        return line, line, line

    def default(self, line):
        """ Add this onto the command string """
        if line.strip() == '!!': return True
        self.command_string += line + '\n'

# To pretty-print usage statements. Some are getting long, so they need to be
# shortened and printed in a helpfully formatted way
def _fmt_usage(usage):
    MAX_LINE_LEN = 79

    def split_options(string):
        whitespace = '\r\n\t '
        openbraces = '[<'
        closebraces = '>]'
        # Splits up the options to keep <> and []'s together
        words, word = [], ''
        i = 0
        # Keep track of when we are inside a <> and when we're not
        nopen = 0
        while i < len(string):
            char = string[i]
            if char in whitespace:
                # Ignore consecutive whitespace
                if not word:
                    i += 1
                    continue
                else:
                    # Only split the string here if we are not in open braces
                    if nopen > 0:
                        word = word.rstrip() + ' '
                        i += 1
                        continue
                    else:
                        words.append(word)
                        word = ''
                        i += 1
                        continue
            # This is not whitespace. If it is an open brace, increment nopen.
            # If it is a closing brace, decrement nopen (check that it does not
            # go below zero)
            if char in openbraces:
                nopen += 1
                word += char
                i += 1
                continue
            if char in closebraces:
                nopen = max(nopen - 1, 0)
                word += char
                i += 1
                continue
            # It is none of these, simply add the character to the word and
            # continue on with the next letter
            word += char
            i += 1
        # Now add the final word to the list of words if it's not blank
        if word: words.append(word)
        return words

    uwords = split_options(usage)
    indent_size = len(uwords[0]) + 1
    lines = []
    line = ''
    for i, uword in enumerate(uwords):
        if not line:
            line = uword
        elif len(line) + len(uword) < MAX_LINE_LEN:
            line += ' ' + uword
        else:
            lines.append(line)
            line = ' ' * indent_size + uword
    if line: lines.append(line)
    return '\n'.join(lines)



"""
This package contains functionality necessary to carry out basic molecular
modelling tasks.
"""

__author__ = 'Jason Swails <jason.swails@gmail.com>'
__date__ = '2015'
__all__ = ['AmberOFFLibrary', 'StandardBiomolecularResidues']

from parmed.modeller.residue import *
from parmed.amber.offlib import AmberOFFLibrary
from parmed.modeller.standardtemplates import StandardBiomolecularResidues

__all__ += residue.__all__



"""
This module contains all of the exceptions that are used in the chemistry
package
"""
import warnings

# Hard errors

class ParmedError(Exception):
    """ Base class for all non-trivial exceptions raised by ParmEd """

class ParsingError(ParmedError):
    """ If there was a problem parsing any kind of file """

class PDBError(ParsingError):
    """ If there was a problem parsing a PDB file """

class Mol2Error(ParsingError):
    """ If there was a problem parsing a Mol2 file """

class MaskError(ParmedError):
    """ Error when a Mask is poorly formed """

class OpenMMError(ParmedError):
    """ If there's a problem making an OpenMM system """

class AmberError(ParsingError):
    """ This is a generic AmberParmError """

class TinkerError(ParsingError):
    """ Raised when one of the TINKER parsing routines hits a bad file """

class CharmmError(ParsingError):
    """ If there is a problem parsing CHARMM PSF files """

class ResidueError(ParmedError):
    """ For when there are problems defining a residue """

class IncompatiblePatchError(ParmedError):
    """ For when applying a PatchTemplate to a ResidueTemplate fails """

class ParameterError(ParmedError):
    """ If a parameter is missing from a database """

class GromacsError(ParmedError):
    """ If there is a problem parsing GROMACS topology files """

class FormatNotFound(ParmedError):
    """ If the file format does not have a registered parser with it """

class RosettaError(ParmedError):
    """ If there is a problem loading a Rosetta pose object """

class PreProcessorError(ParmedError):
    """ If there is a problem running the C-like preprocessor """

class MoleculeError(ParmedError):
    """ If there is a problem defining a molecule via the bond graph """

class PdbxError(ParmedError):
    """ Class for catching general errors with PDBx/mmCIF parsing """

class PdbxSyntaxError(PdbxError):
    """ Class for catching errors in mmCIF/PDBx syntax """
    def __init__(self, lineNumber='-1', text=''):
        Exception.__init__(self)
        self.lineNumber = lineNumber
        self.text = text

    def __str__(self):
        return "%%ERROR - [at line: %d] %s" % (self.lineNumber, self.text)

class InputError(ParmedError):
    """ When there is an error with input """

# Warnings

class ParmedWarning(Warning):
    """ Base class for all warnings raised by ParmEd """

# Make sure that all warnings are always printed
warnings.filterwarnings('always', category=ParmedWarning)

class PDBWarning(ParmedWarning):
    """ A non-fatal error to indicate a problematic PDB file """

class AmberWarning(ParmedWarning):
    """ If there is something that is non-fatal """

class SplitResidueWarning(ParmedWarning):
    """ For if a residue with the same number but different names is split """

class CharmmWarning(ParmedWarning):
    """ For non-fatal PSF parsing issues """

class GromacsWarning(ParmedWarning):
    " If we are uncertain about something regarding the GROMACS topology file "

class ParameterWarning(ParmedWarning):
    """ If a type of parameter is missing, but you don't want it to be fatal """

class TinkerWarning(ParmedWarning):
    pass

class PreProcessorWarning(ParmedWarning):
    """ If there is something we should warn about in preprocessing """

class OpenMMWarning(ParmedWarning):
    """ If there is something we should warn when processing OpenMM objects """

# Control flow exceptions

class CharmmPsfEOF(ParmedError):
    """ If we hit an end-of-file in parsing CHARMM files """



"""
This module contains classes for reading various TINKER-style files
"""
from parmed.exceptions import TinkerError
from parmed.formats.registry import FileFormatType, load_file
from parmed.periodic_table import element_by_name, AtomicNum
from parmed.structure import Structure
from parmed.topologyobjects import Atom, Bond, Residue
from parmed.utils.io import genopen
from parmed.utils.six import add_metaclass, string_types
from parmed.utils.six.moves import range, zip

class KeywordControlFile(object):
    """ Reads and processes a keyword control file for TINKER simulations """
    _datatypes = {'PARAMETERS' : str, 'A-AXIS' : float, 'B-AXIS' : float,
                  'C-AXIS' : float, 'ALPHA' : float, 'BETA' : float,
                  'GAMMA' : float}

    def __init__(self, fname):
        # The currently recognized/parsed keywords (these are the only ones
        # necessary for running Amoeba in Amber)
        self.keywords = {'PARAMETERS' : None, 'A-AXIS' : None, 'B-AXIS' : None,
                         'C-AXIS' : None, 'ALPHA' : None, 'BETA' : None,
                         'GAMMA' : None}
        # Parse the file
        for line in open(fname, 'r'):
            # Skip over any blank lines
            if not line.strip(): continue
            words = line.split()
            key = words[0].upper()
            # Get rid of the keyword and all whitespace
            result = line.replace(words[0], '').strip()
            try:
                self.keywords[key] = self._datatypes[key](result)
            except KeyError:
                pass # All non-keyword lines are ignored comments
            except ValueError:
                raise TinkerError('Malformed keyword control file! Could not '
                                  'convert the value of %s into type %s' %
                                  (key, self._datatypes[key]))

@add_metaclass(FileFormatType)
class XyzFile(Structure):
    """ Reads and processes a Tinker XYZ file

    Parameters
    ----------
    fname : str or file-like
        Name of the file, or the file object containing the XYZ file contents
    seq : str, optional
        Name of the file containing the residue (and chain) sequence. Default is
        None (so every atom will be part of the same residue)
    """

    @staticmethod
    def _check_atom_record(words):
        """ Checks that the tokenized line is consistent with an atom record

        Parameters
        ----------
        words : list of str
            Result of line.split() on a line of the file

        Returns
        -------
        is_atom : bool
            If it is consistent with an atom record, return True. Otherwise,
            False
        """
        if len(words) < 6: return False
        # First token is the atom index
        try:
            int(words[0])
        except ValueError:
            return False
        # Second token is the atom name (NOT numeric)
        try:
            float(words[1])
        except ValueError:
            try:
                int(words[1])
            except ValueError:
                pass
            else:
                return False
        else:
            return False
        # Third, fourth, and fifth tokens are floats, but NOT integers
        try:
            [float(w) for w in words[2:5]]
        except ValueError:
            return False
        else:
            try:
                [int(w) for w in words[2:5]]
            except ValueError:
                pass
            else:
                return False
        # Remaining numbers are integers
        try:
            [int(w) for w in words[5:]]
        except ValueError:
            return False
        return True

    @staticmethod
    def id_format(filename):
        """ Identify the file as a Tinker XYZ file

        Parameters
        ----------
        filename : str
            Name of the file to test whether or not it is a mol2 file

        Returns
        -------
        is_fmt : bool
            True if it is a xyz file, False otherwise
        """
        f = genopen(filename, 'r')
        words = f.readline().split() # natom and title
        if not words:
            return False
        try:
            natom = int(words[0])
        except (ValueError, IndexError):
            return False
        else:
            if natom <= 0:
                return False
        words = f.readline().split()
        # Either a box line or a line with the first atom
        if len(words) == 6:
            try:
                [float(w) for w in words]
            except ValueError:
                if XyzFile._check_atom_record(words):
                    return True
                return False
            else:
                # Next line needs to be an atom record
                words = f.readline().split()
        return XyzFile._check_atom_record(words)

    def __init__(self, fname, seq=None):
        super(XyzFile, self).__init__()
        if isinstance(fname, string_types):
            fxyz = genopen(fname, 'r')
            own_handle_xyz = True
        else:
            fxyz = fname
            own_handle_xyz = False
        if seq is not None:
            seqstruct = load_file(seq)
        # Now parse the file
        try:
            natom = int(fxyz.readline().split()[0])
        except (ValueError, IndexError):
            raise TinkerError('Bad XYZ file format; first line')
        if seq is not None and natom != len(seqstruct.atoms):
            raise ValueError('Sequence file %s # of atoms does not match the # '
                             'of atoms in the XYZ file' % seq)
        words = fxyz.readline().split()
        if len(words) == 6 and not XyzFile._check_atom_record(words):
            self.box = [float(w) for w in words]
            words = fxyz.readline().split()
        residue = Residue('SYS')
        residue.number = 1
        residue._idx = 0
        if seq is not None:
            residue = seqstruct.residues[0]
            atomic_number = _guess_atomic_number(words[1], residue)
        else:
            atomic_number = AtomicNum[element_by_name(words[1])]
        atom = Atom(atomic_number=atomic_number, name=words[1], type=words[5])
        atom.xx, atom.xy, atom.xz = [float(w) for w in words[2:5]]
        self.add_atom(atom, residue.name, residue.number, residue.chain,
                      residue.insertion_code, residue.segid)
        bond_ids = [[int(w) for w in words[6:]]]
        for i, line in enumerate(fxyz):
            words = line.split()
            if seq is not None:
                residue = seqstruct.atoms[i+1].residue
                atomic_number = _guess_atomic_number(words[1], residue)
            else:
                atomic_number = AtomicNum[element_by_name(words[1])]
            atom = Atom(atomic_number=atomic_number, name=words[1],
                        type=words[5])
            atom.xx, atom.xy, atom.xz = [float(w) for w in words[2:5]]
            self.add_atom(atom, residue.name, residue.number, residue.chain,
                          residue.insertion_code, residue.segid)
            bond_ids.append([int(w) for w in words[6:]])
        # All of the bonds are stored now -- go ahead and make them now
        for atom, bonds in zip(self.atoms, bond_ids):
            i = atom.idx + 1
            for idx in bonds:
                if idx > i:
                    self.bonds.append(Bond(atom, self.atoms[idx-1]))
        if seq is None:
            # Try to improve atomic number prediction for monoatomic species
            # (like ions) if no sequence as loaded
            for atom in self.atoms:
                if len(atom.bonds) == 0: # not bonded to anybody else
                    atom.atomic_number = _guess_atomic_number(atom.name)
        if own_handle_xyz:
            fxyz.close()

class DynFile(object):
    """ Reads and processes a Tinker DYN file """
    def __init__(self, fname=None):
        if fname is not None:
            self.read(fname)

    def read(self, fname):
        """ Parses the .dyn file """
        f = open(fname, 'r')
        try:
            if f.readline().strip() != 'Number of Atoms and Title :':
                raise TinkerError('%s is not a recognized TINKER .dyn file' %
                                  fname)
            line = f.readline()
            self.natom, self.title = int(line[0:8].strip()), line[8:].strip()
            if f.readline().strip() != 'Periodic Box Dimensions :':
                raise TinkerError('No periodic box dimension line')
            self.box = [0.0 for i in range(6)]
            words = f.readline().upper().split()
            self.box[0] = float(words[0].replace('D', 'E'))
            self.box[1] = float(words[1].replace('D', 'E'))
            self.box[2] = float(words[2].replace('D', 'E'))
            words = f.readline().upper().split()
            self.box[3] = float(words[0].replace('D', 'E'))
            self.box[4] = float(words[1].replace('D', 'E'))
            self.box[5] = float(words[2].replace('D', 'E'))
            if f.readline().strip() != 'Current Atomic Positions :':
                raise TinkerError('No atomic positions in .dyn file')
            self.positions = [[0.0, 0.0, 0.0] for i in range(self.natom)]
            DynFile._read_section(f, self.positions, self.natom)
            line = f.readline().strip()
            if line == 'Current Translational Velocities :':
                self.rigidbody = True
                self.translational_velocities = [[0.0, 0.0, 0.0] for i in
                            range(self.natom)]
                DynFile._read_section(f, self.translational_velocities,
                                      self.natom)
                if f.readline().strip() != 'Current Angular Velocities :':
                    raise TinkerError('Could not find Angular velocity '
                                      'section in .dyn file')
                self.angular_velocities = [[0.0, 0.0, 0.0]
                                            for i in range(self.natom)]
                DynFile._read_section(f, self.angular_velocities, self.natom)
                if f.readline().strip() != 'Current Angular Momenta :':
                    raise TinkerError('Could not find angular momenta section '
                                      'in .dyn file')
                self.angular_momenta = [[0.0, 0.0, 0.0]
                                        for i in range(self.natom)]
                DynFile._read_section(f, self.angular_momenta, self.natom)
            elif line == 'Current Atomic Velocities :':
                self.rigidbody = False
                self.velocities = [[0.0, 0.0, 0.0] for i in range(self.natom)]
                DynFile._read_section(f, self.velocities, self.natom)
                if f.readline().strip() != 'Current Atomic Accelerations :':
                    raise TinkerError('Could not find accelerations in %s ' %
                                      fname)
                self.accelerations = [[0.0, 0.0, 0.0]
                                      for i in range(self.natom)]
                DynFile._read_section(f, self.accelerations, self.natom)
                if f.readline().strip() != 'Alternate Atomic Accelerations :':
                    raise TinkerError('Could not find old accelerations in %s' %
                                      fname)
                self.old_accelerations = [[0.0, 0.0, 0.0]
                                          for i in range(self.natom)]
                DynFile._read_section(f, self.old_accelerations, self.natom)
            else:
                raise TinkerError('No velocities in %s' % fname)
        finally:
            f.close()

    @staticmethod
    def _read_section(f, container, natom):
        """
        Reads a section of an open file into the passed container. The
        container should be a 2-dimensional list of length natom x 3
        """
        for i in range(natom):
            words = f.readline().upper().split()
            try:
                container[i][0] = float(words[0].replace('D', 'E'))
                container[i][1] = float(words[1].replace('D', 'E'))
                container[i][2] = float(words[2].replace('D', 'E'))
            except (IndexError, ValueError):
                raise TinkerError('Could not parse values from dyn file')

def is_float(thing):
    try:
        float(thing)
        return True
    except ValueError:
        return False

def _guess_atomic_number(name, residue=None):
    """ Guesses the atomic number """
    # Special-case single-atom residues, which are almost always ions
    name = ''.join(c for c in name if c.isalpha())
    if residue is None or len(residue.atoms) == 1:
        if len(name) > 1:
            try:
                return AtomicNum[name[0].upper() + name[1].lower()]
            except KeyError:
                return AtomicNum[element_by_name(name)]
    return AtomicNum[element_by_name(name)]



#!/bin/env python
"""
Module simtk.unit.unit_definitions

This is part of the OpenMM molecular simulation toolkit originating from
Simbios, the NIH National Center for Physics-Based Simulation of
Biological Structures at Stanford, funded under the NIH Roadmap for
Medical Research, grant U54 GM072970. See https://simtk.org.

Portions copyright (c) 2012 Stanford University and the Authors.
Authors: Christopher M. Bruns
Contributors: Peter Eastman

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import division, print_function, absolute_import

__author__ = "Christopher M. Bruns"
__version__ = "0.6"

from .baseunit import BaseUnit
from .standard_dimensions import *
from .unit import Unit, ScaledUnit, UnitSystem, dimensionless
from .unit_operators import * ; # needed for manipulation of units
from .prefix import *
import math
import sys

#####################
### DIMENSIONLESS ###
#####################

# dimensionless = Unit({}); # defined in unit.py

##############
### LENGTH ###
##############

meter_base_unit = BaseUnit(length_dimension, "meter", "m")
meters = meter = Unit({meter_base_unit: 1.0})
define_prefixed_units(meter_base_unit, module = sys.modules[__name__])

angstrom_base_unit = BaseUnit(length_dimension, "angstrom", "A")
angstrom_base_unit.define_conversion_factor_to(meter_base_unit, 1e-10)
angstroms = angstrom = Unit({angstrom_base_unit: 1.0})

planck_length_base_unit = BaseUnit(length_dimension, "Planck length", "l_P")
planck_length_base_unit.define_conversion_factor_to(meter_base_unit, 1.61625281e-35)

inch_base_unit = BaseUnit(length_dimension, "inch", "in")
inch_base_unit.define_conversion_factor_to(centimeter_base_unit, 2.5400)
inch = inches = Unit({inch_base_unit: 1.0})

foot_base_unit = BaseUnit(length_dimension, "foot", "ft")
foot_base_unit.define_conversion_factor_to(inch_base_unit, 12.0)
foot = feet = Unit({foot_base_unit: 1.0})

yard_base_unit = BaseUnit(length_dimension, "yard", "yd")
yard_base_unit.define_conversion_factor_to(foot_base_unit, 3.0)
yard = yards = Unit({yard_base_unit: 1.0})

furlongs = furlong = yard.create_unit(scale=220.0, name="furlong", symbol="furlong")
miles = mile = furlong.create_unit(scale=8.0, name="mile", symbol="mi")

############
### MASS ###
############

gram_base_unit = BaseUnit(mass_dimension, "gram", "g")
grams = gram = Unit({gram_base_unit: 1.0})

define_prefixed_units(gram_base_unit, module = sys.modules[__name__])

planck_mass_base_unit = BaseUnit(mass_dimension, "Planck mass", "m_P")
planck_mass_base_unit.define_conversion_factor_to(kilogram_base_unit, 2.1764411e-8)

# pound can be mass, force, or currency
pound_mass_base_unit = BaseUnit(mass_dimension, "pound", "lb")
pound_mass_base_unit.define_conversion_factor_to(kilogram_base_unit, 0.3732)
pound_mass = pounds_mass = Unit({pound_mass_base_unit: 1.0})

stone_base_unit = BaseUnit(mass_dimension, "stone", "stone")
stone_base_unit.define_conversion_factor_to(pound_mass_base_unit, 14.0)
stone = stones = Unit({stone_base_unit: 1.0})

############
### TIME ###
############

second_base_unit = BaseUnit(time_dimension, "second", "s")
seconds = second = Unit({second_base_unit: 1.0})

define_prefixed_units(second_base_unit, module = sys.modules[__name__])

planck_time_base_unit = BaseUnit(time_dimension, "Planck time", "t_P")
planck_time_base_unit.define_conversion_factor_to(second_base_unit, 5.3912427e-44)

minutes = minute = second.create_unit(scale=60.0, name="minute", symbol="min")
hours = hour = minute.create_unit(scale=60.0, name="hour", symbol="hr")
days = day = hour.create_unit(scale=24.0, name="day", symbol="day")
weeks = week = day.create_unit(scale=7.0, name="week", symbol="week")
years = year = day.create_unit(scale=365.25, name="julian year", symbol="a")
centuries = centurys = century = year.create_unit(scale=100.0, name="century", symbol="century")
millenia = milleniums = millenium = century.create_unit(scale=10.0, name="millenium", symbol="ka")

fortnights = fortnight = day.create_unit(scale=14.0, name="fortnight", symbol="fortnight")

###################
### TEMPERATURE ###
###################

kelvin_base_unit = BaseUnit(temperature_dimension, "kelvin", "K")
kelvins = kelvin = Unit({kelvin_base_unit: 1.0})

planck_temperature_base_unit = BaseUnit(temperature_dimension, "Planck temperature", "T_p")
planck_temperature_base_unit.define_conversion_factor_to(kelvin_base_unit, 1.41678571e32)

##############
### CHARGE ###
##############

elementary_charge_base_unit = BaseUnit(charge_dimension, "elementary charge", "e")
elementary_charges = elementary_charge = Unit({elementary_charge_base_unit: 1.0})

coulomb_base_unit = BaseUnit(charge_dimension, "coulomb", "C")
# Exact conversion factor
coulomb_base_unit.define_conversion_factor_to(elementary_charge_base_unit, 6.24150962915265e18)
coulombs = coulomb = Unit({coulomb_base_unit: 1.0})

planck_charge_base_unit = BaseUnit(charge_dimension, "Planck charge", "q_P")
planck_charge_base_unit.define_conversion_factor_to(elementary_charge_base_unit, 11.706237639840)

##############
### AMOUNT ###
##############

mole_base_unit = BaseUnit(amount_dimension, "mole", "mol")
moles = mole = Unit({mole_base_unit: 1.0})

single_item_amount_base_unit = BaseUnit(amount_dimension, "item", "")
mole_base_unit.define_conversion_factor_to(single_item_amount_base_unit, 6.0221417930e23)
items = item = Unit({single_item_amount_base_unit: 1.0})

##########################
### Luminous Intensity ###
##########################

candela_base_unit = BaseUnit(luminous_intensity_dimension, "candela", "cd")
candelas = candela = Unit({candela_base_unit: 1.0})

#############
### ANGLE ###
#############

radian_base_unit = BaseUnit(angle_dimension, "radian", "rad")
radians = radian = Unit({radian_base_unit: 1.0})

degree_base_unit = BaseUnit(angle_dimension, "degree", "deg")
degree_base_unit.define_conversion_factor_to(radian_base_unit, math.pi/180.0)
degrees = degree = Unit({degree_base_unit: 1.0})

arcminutes = arcminute = degree.create_unit(scale=1.0/60.0, name="arcminute", symbol="'")
arcseconds = arcsecond = arcminute.create_unit(scale=1.0/60.0, name="arcsecond", symbol='"')

###################
### INFORMATION ###
###################

bit_base_unit = BaseUnit(information_dimension, "bit", "bit")
bits = bit = Unit({bit_base_unit: 1.0})

byte_base_unit = BaseUnit(information_dimension, "byte", "B")
byte_base_unit.define_conversion_factor_to(bit_base_unit, 8.0)
bytes = byte = Unit({byte_base_unit: 1.0})

nat_base_unit = BaseUnit(information_dimension, "nat", "nat")
nat_base_unit.define_conversion_factor_to(bit_base_unit, 1.0/math.log(2.0))
nats = nat = nits = nit = nepits = nepit = Unit({nat_base_unit: 1.0})

ban_base_unit = BaseUnit(information_dimension, "ban", "ban")
ban_base_unit.define_conversion_factor_to(bit_base_unit, math.log(10.0, 2.0))
bans = ban = hartleys = hartley = dits = dit = Unit({ban_base_unit: 1.0})


###############
### DERIVED ###
###############

# Molar mass
# daltons = dalton = grams / mole
daltons = dalton = Unit({ScaledUnit(1.0, gram/mole, "dalton", "Da"): 1.0})
amus = amu = dalton
atom_mass_units = atomic_mass_unit = dalton

# Volume
liter_base_unit = ScaledUnit(1.0, decimeter**3, "liter", "L")
liter = liters = litre = litres = Unit({liter_base_unit: 1.0})
define_prefixed_units(liter_base_unit, module = sys.modules[__name__])

# Concentration
molar_base_unit = ScaledUnit(1.0, mole/liter, "molar", "M")
molar = molal = Unit({molar_base_unit: 1.0})
define_prefixed_units(molar_base_unit, module = sys.modules[__name__])

# Force
newton_base_unit = ScaledUnit(1.0, kilogram * meter / second / second, "newton", "N")
newtons = newton = Unit({newton_base_unit: 1.0})
define_prefixed_units(newton_base_unit, module = sys.modules[__name__])
# pound can be mass, force, or currency
pound_force_base_unit = ScaledUnit(4.448, newton, "pound", "lb")
pound_force = pounds_force = Unit({pound_force_base_unit: 1.0})
dyne_base_unit = ScaledUnit(1.0, gram * centimeter / second**2, "dyne", "dyn")
dyne = dynes = Unit({dyne_base_unit: 1.0})

# Energy
joule_base_unit = ScaledUnit(1.0, newton * meter, "joule", "J")
joules = joule = Unit({joule_base_unit: 1.0})
define_prefixed_units(joule_base_unit, module = sys.modules[__name__])
erg_base_unit = ScaledUnit(1.0, dyne * centimeter, "erg", "erg")
erg = ergs = Unit({erg_base_unit: 1.0})

# In molecular simulations, "kilojoules" are in microscopic units
# And you really only want to use kilojoules/mole.
md_kilojoule_raw = gram * nanometer**2 / picosecond**2
md_kilojoules = md_kilojoule = Unit({ScaledUnit(1.0, md_kilojoule_raw, "kilojoule", "kJ"): 1.0})
kilojoules_per_mole = kilojoule_per_mole = md_kilojoule / mole

calorie_base_unit = ScaledUnit(4.184, joule, "calorie", "cal")
calories = calorie = Unit({calorie_base_unit: 1.0})
define_prefixed_units(calorie_base_unit, module = sys.modules[__name__])
md_kilocalories = md_kilocalorie = Unit({ScaledUnit(4.184, md_kilojoule, "kilocalorie", "kcal"): 1.0})
kilocalories_per_mole = kilocalorie_per_mole = md_kilocalorie / mole

# Power
watt_base_unit = ScaledUnit(1.0, joule / second, "watt", "W")
watt = watts = Unit({watt_base_unit: 1.0})

# Current
ampere_base_unit = ScaledUnit(1.0, coulomb / second, "ampere", "A")
ampere = amperes = amp = amps = Unit({ampere_base_unit: 1.0})

# Electrical potential
volt_base_unit = ScaledUnit(1.0, watt / ampere, "volt", "V")
volt = volts = Unit({volt_base_unit: 1.0})

# Magnetic field
tesla_base_unit = ScaledUnit(1.0, newton / (ampere * meter), "tesla", "T")
tesla = teslas = Unit({tesla_base_unit: 1.0})
gauss_base_unit = ScaledUnit(10.0**-4, tesla, "gauss", "G")
gauss = Unit({gauss_base_unit: 1.0})

# Electrical resistance
ohm_base_unit = ScaledUnit(1.0, volt / ampere, "ohm", "O")
ohm = ohms = Unit({ohm_base_unit: 1.0})

# Capacitance
farad_base_unit = ScaledUnit(1.0, coulomb / volt, "farad", "F")
farad = farads = Unit({farad_base_unit: 1.0})

# Inductance
henry_base_unit = ScaledUnit(1.0, volt * second / ampere, "henry", "H")
henry = henries = henrys = Unit({henry_base_unit: 1.0})

# Polarizability
debye_base_unit = ScaledUnit(0.20822678, elementary_charge * angstrom, "debye", "D")
debyes = debye = Unit({debye_base_unit: 1.0})

# Pressure
pascal_base_unit = ScaledUnit(1.0, newton / (meter**2), "pascal", "Pa")
pascals = pascal = Unit({pascal_base_unit: 1.0})
define_prefixed_units(pascal_base_unit, module = sys.modules[__name__])
psi_base_unit = ScaledUnit(1.0, pound_force / (inch**2), "psi", "psi")
psi = Unit({psi_base_unit: 1.0})
bar_base_unit = ScaledUnit(10.0**5, pascal, "bar", "bar")
bar = bars = Unit({bar_base_unit: 1.0})
atmosphere_base_unit = ScaledUnit(101325.0, pascal, "atmosphere", "atm")
atmosphere = atmospheres = Unit({atmosphere_base_unit: 1.0})
torr_base_unit = ScaledUnit(1.0/760.0, atmosphere, "torr", "Torr")
torr = Unit({torr_base_unit: 1.0})
mmHg_base_unit = ScaledUnit(133.322, pascal, "mmHg", "mmHg")
mmHg = Unit({mmHg_base_unit: 1.0})

####################
### Unit Systems ###
####################

ampere_base_unit = ScaledUnit(1.0, coulomb/second, "ampere", "A")

si_unit_system = UnitSystem([
        meter_base_unit,
        kilogram_base_unit,
        second_base_unit,
        ampere_base_unit,
        kelvin_base_unit,
        mole_base_unit,
        candela_base_unit,
        radian_base_unit])

cgs_unit_system = UnitSystem([
        centimeter_base_unit,
        gram_base_unit,
        second_base_unit,
        ampere_base_unit,
        kelvin_base_unit,
        mole_base_unit,
        radian_base_unit])

dalton_base_unit = ScaledUnit(1.0, gram/mole, "dalton", "Da")

md_unit_system = UnitSystem([
        nanometer_base_unit,
        dalton_base_unit,
        picosecond_base_unit,
        elementary_charge_base_unit,
        kelvin_base_unit,
        mole_base_unit,
        radian_base_unit])

planck_unit_system = UnitSystem([\
        planck_length_base_unit,
        planck_mass_base_unit,
        planck_time_base_unit,
        planck_charge_base_unit,
        planck_temperature_base_unit,
        single_item_amount_base_unit,
        radian_base_unit])

########################
### TESTING/EXAMPLES ###
########################

# run module directly for testing
if __name__=='__main__':
    # Test the examples in the docstrings
    import doctest, sys
    doctest.testmod(sys.modules[__name__])



"""
This module will create an amber mdin file for either sander or      
pmemd (or others). The program specification loads the appropriate   
dictionaries with default values, etc. It can read and write mdins. 

                           GPL LICENSE INFO                             

Copyright (C) 2009 - 2014 Dwight Mcgee, Bill Miller III, and Jason Swails

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
   
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
"""

# This module will create and read a sander/pmemd input
from parmed.amber.mdin.cntrl import cntrl
from parmed.amber.mdin.ewald import ewald
from parmed.amber.mdin.pb import pb
from parmed.amber.mdin.qmmm import qmmm
from parmed.exceptions import InputError
from parmed.utils.six import string_types
from parmed.utils.six.moves import range

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def addOn(line, string, file):
    if len(line.strip()) == 0:
        return line + string
    elif len(line) + len(string) > 40:
        file.write(line + '\n')
        return ' ' + string
    else:
        return line + string

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Mdin(object):
   
   #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, program = 'sander', verbosity = 1):
        # define instance data
        self.program = program   # which program we're creating the input file for
        self.cntrl_obj = cntrl() # object with cntrl namelist vars in a dictionary
        self.ewald_obj = ewald() # object with ewald namelist vars in a dictionary
        self.pb_obj = pb()       # object with pb namelist vars in a dictionary
        self.qmmm_obj = qmmm()   # object with qmmm namelist vars in a dictionary
        self.verbosity = 0       # verbosity level: 0 -- print nothing
                                 #                  1 -- print errors
                                 #                  2 -- 1 + warnings
                                 #                  3 -- 2 + notes
        self.cards = []          # array that has all of the input cards that come
                                 # after namelists
        self.cntrl_nml = {}      # dictionary with cntrl namelist vars
        self.cntrl_nml_defaults = {} # dictionary with default cntrl namelist vars
        self.ewald_nml = {}          # dictionary with ewald namelist vars
        self.ewald_nml_defaults = {} # dictionary with default ewald namelist vars
        self.pb_nml = {}             # dictionary with pb namelist vars
        self.pb_nml_defaults = {}    # dictionary with default pb namelist vars
        self.qmmm_nml = {}           # dictionary with qmmm namelist vars
        self.qmmm_nml_defaults = {}  # dictionary with default qmmm namelist vars
        self.valid_namelists = []    # array with valid namelists for each program
        self.title = 'mdin prepared by mdin.py'   # title for the mdin file


        if self.program == "sander":
            self.cntrl_nml = self.cntrl_obj.sander
            self.ewald_nml = self.ewald_obj.sander
            self.pb_nml = self.pb_obj.sander
            self.qmmm_nml = self.qmmm_obj.sander
            self.valid_namelists = ['cntrl','ewald','qmmm','pb']
        elif self.program == "sander.APBS":
            self.cntrl_nml = self.cntrl_obj.sander
            self.pb_nml = self.pb_obj.sanderAPBS
            self.valid_namelists = ['cntrl','apbs']
        elif self.program == "pmemd":
            self.cntrl_nml = self.cntrl_obj.pmemd
            self.ewald_nml = self.ewald_obj.pmemd
            self.valid_namelists = ['cntrl','ewald']
        else:
            raise InputError('Error: program (%s) unrecognized!' %
                                   self.program)

        self.cntrl_nml_defaults = self.cntrl_nml.copy()
        self.ewald_nml_defaults = self.ewald_nml.copy()
        self.pb_nml_defaults = self.pb_nml.copy()
        self.qmmm_nml_defaults = self.qmmm_nml.copy()

   #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def write(self, filename = 'mdin'):
      
        # open the file for writing and write the header and &cntrl namelist
        file = open(filename,'w')
        file.write(self.title + '\n')
        file.write('&cntrl\n')
        # automatic indent of single space
        line = ' '
        # add any variable that is different from the default to the mdin file
        for var in self.cntrl_nml.keys():
            if self.cntrl_nml[var] == self.cntrl_nml_defaults[var]: continue
            if isinstance(self.cntrl_nml[var], string_types):
                line = addOn(line, "%s='%s', " % (var, self.cntrl_nml[var]), file)
            else:
                line = addOn(line, "%s=%s, " % (var, self.cntrl_nml[var]), file)

        # flush any remaining items that haven't yet been printed to the mdin file
        if len(line.strip()) != 0:
            file.write(line + '\n')

        # end the namelist
        file.write('/\n')

        # print the ewald namelist if any variables differ from the default
        line = ' '
        has_been_printed = False  # keep track if this namelist has been printed
        for var in self.ewald_nml.keys():
            if self.ewald_nml[var] == self.ewald_nml_defaults[var]: continue
            if not has_been_printed:
                file.write('&ewald\n')
                has_been_printed = True
            if isinstance(self.ewald_nml_defaults[var], string_types):
                line = addOn(line, "%s='%s', " % (var, self.ewald_nml[var]), file)
            else:
                line = addOn(line, "%s='%s', " % (var, self.ewald_nml[var]), file)

        # flush any remaining items that haven't been printed to the mdin file
        if len(line.strip()) != 0:
            file.write(line + '\n')

        # end the namelist
        if has_been_printed:
            file.write('/\n')

        # print the pb namelist if any variables differ from the original
        line = ' '
        has_been_printed = False # keep track if this namelist has been printed
        for var in self.pb_nml.keys():
            if self.pb_nml[var] == self.pb_nml_defaults[var]: continue
            if not has_been_printed:
                if self.program == 'sander.APBS':
                    file.write('&apbs\n')
                else:
                    file.write('&pb\n')
                has_been_printed = True
            if isinstance(self.pb_nml[var], string_types):
                line = addOn(line,"%s='%s', " % (var, self.pb_nml[var]), file)
            else:
                line = addOn(line,"%s=%s, " % (var, self.pb_nml[var]), file)

        # flush any remaining items that haven't been printed to the mdin file
        if len(line.strip()) != 0:
            file.write(line + '\n')

        # end the namelist
        if has_been_printed:
            file.write('/\n')

        # print the qmmm namelist if any variables differ from the original
        line = ' '
        has_been_printed = False # keep track if this namelist has been printed
        if self.cntrl_nml['ifqnt'] == 1:
            for var in self.qmmm_nml.keys():
                if self.qmmm_nml[var] == self.qmmm_nml_defaults[var]: continue
                if not has_been_printed:
                    file.write('&qmmm\n')
                    has_been_printed = True
                if isinstance(self.qmmm_nml_defaults[var], string_types):
                    line = addOn(line, "%s='%s', " % (var, self.qmmm_nml[var]),
                                 file)
                else:
                    line = addOn(line, "%s=%s, " % (var, self.qmmm_nml[var]), file)
      
            # flush any remaining items that haven't been printed to the mdin file
            if len(line.strip()) != 0:
                file.write(line + '\n')

            # end the namelist
            if has_been_printed:
                file.write('/\n')

        # Write the cards to the input file
        for i in range(len(self.cards)):
            file.write(self.cards[i].strip() + '\n')
        if len(self.cards) != 0:
            file.write('END\n')
      
        file.close()

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
   
    def read(self, filename = 'mdin'):
        lines = open(filename, 'r').readlines()

        # split up input file into separate fields by comma
        blocks = []  # namelists in the order they appear
        block_fields = [] # array of arrays that correspond to entries in 
                          # namelists found in "blocks" above
        inblock = False
        lead_comment = True
        for i in range(len(lines)):
            if not inblock and not lines[i].strip().startswith('&') and \
                    lead_comment:
                continue
            elif not inblock and not lines[i].strip().startswith('&') and \
                    not lead_comment:
                final_ended = True
                for j in range(i,len(lines)):
                    if lines[j].strip().startswith('&'):
                        final_ended = False
                if final_ended and len(lines[i].strip()) != 0:
                    self.cards.append(lines[i])
            elif not inblock and lines[i].strip().startswith('&'):
                lead_comment = False
                inblock = True
                block = lines[i].strip()[1:].lower()
                blocks.append(block)    # add the name of the namelist to "blocks"
                block_fields.append([]) # add empty array to be filled with entries 
                                        # for given namelist
                if not block in self.valid_namelists:
                    raise InputError('Invalid namelist (%s) in input '
                            'file (%s) for %s' % (lines[i].strip(), filename,
                            self.program))
            elif inblock and (lines[i].strip() == '/' or 
                            lines[i].strip() == '&end'):
                inblock = False
            elif inblock and lines[i].strip().startswith('&'):
                raise InputError('Invalid input file (%s). Terminate each '
                                'namelist before another is started' % filename)
            elif inblock:
                items = lines[i].strip().split(',')
                j = 0
                while j < len(items):
                    items[j] = items[j].strip()
                    if len(items[j]) == 0:
                        items.pop(j)
                    else:
                        j += 1
                block_fields[len(block_fields)-1].extend(items)

        # take out the last END in the cards if it's there
        if len(self.cards) != 0 and \
                self.cards[len(self.cards)-1].strip().upper() == 'END':
            self.cards.pop()

        # combine any multi-element fields: e.g. rstwt=1,2,3,
        begin_field = -1
        for i in range(len(block_fields)):
            for j in range(len(block_fields[i])):
                if not '=' in block_fields[i][j]:
                    if begin_field == -1:
                        raise InputError('Invalid input file (%s).' %
                                               filename)
                    else:
                        block_fields[i][begin_field] += ',' + block_fields[i][j]
                else:
                    begin_field = j

        # now parse through the options and add them to the dictionaries
        for i in range(len(block_fields)):
            for j in range(len(block_fields[i])):
                if not '=' in block_fields[i][j]:
                    continue
                else:
                    var = block_fields[i][j].split('=')
                    self.change(blocks[i], var[0].strip(), var[1].strip())

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def change(self, namelist, variable, value):
        """ Change the value of a variable without adding a new key-pair """
      
        variable = variable.lower()
        if isinstance(value, string_types):
            if (value.startswith('"') and value.endswith('"')) or (
                   value.startswith("'") and value.endswith("'")):
                value = value[1:-1]
      
        if namelist == "cntrl":
            if variable in self.cntrl_nml.keys():
                mytype = type(self.cntrl_nml_defaults[variable])
                self.cntrl_nml[variable] = mytype(value)
            else:
                raise InputError('Unknown variable (%s) in &cntrl!' % variable)
        elif namelist == 'ewald': 
            if variable in self.ewald_nml.keys():
                mytype = type(self.ewald_nml_defaults[variable])
                self.ewald_nml[variable] = mytype(value)
            else:
                raise InputError('Unknown variable (%s) in &ewald!' % variable)
        elif namelist == 'pb' or namelist == 'apbs':
            if variable in self.pb_nml.keys():
                mytype = type(self.pb_nml_defaults[variable])
                self.pb_nml[variable] = mytype(value)
            else:
                raise InputError('Unknown variable (%s) in &%s!' % 
                                (variable, namelist))
        elif namelist == 'qmmm':
            if variable in self.qmmm_nml.keys():
                mytype = type(self.qmmm_nml_defaults[variable])
                self.qmmm_nml[variable] = mytype(value)
            else:
                raise InputError('Unknown variable (%s) in &qmmm' % variable)
        else:
            raise InputError('Unknown namelist (%s)!' % namelist)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def check(self):
        return True

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def SHAKE(self):
        self.change('cntrl','ntf', 2)
        self.change('cntrl','ntc', 2)
        self.change('cntrl','dt', 0.002)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def constPressure(self, press=1.0, taup=1.0):
        self.change('cntrl','ntb', 2)
        self.change('cntrl','ntp', 1)
        self.change('cntrl','pres0', press)
        self.change('cntrl','taup', taup)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def constVolume(self):
        self.change('cntrl','ntb', 1)
        self.change('cntrl','ntp', 0)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def constTemp(self, ntt=3, temp=300.0, gamma_ln=2.0, ig=-1, tautp=1.0):
        self.change('cntrl','ntt', ntt)
        self.change('cntrl','temp0', temp)
        self.change('cntrl','tempi', temp)
        self.change('cntrl','gamma_ln', gamma_ln if ntt==3 else 0)
        self.change('cntrl','ig', ig)
        self.change('cntrl','tautp', tautp)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def constpH(self, solvph=7.0, igb=2, ntcnstph=10):
        self.change('cntrl','icnstph', 1)
        self.change('cntrl','solvph', solvph)
        self.change('cntrl','ntcnstph', ntcnstph)
        self.change('cntrl','igb', igb)
        self.change('cntrl','ntb', 0)
        self.change('cntrl','saltcon', 0.1)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def restrainHeavyAtoms(self, restraint_wt=0.0):
        self.change('cntrl','ntr', 1)
        self.change('cntrl','restraint_wt', restraint_wt)
        self.change('cntrl','restraintmask', '!@H=')

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def restrainBackbone(self, restraint_wt=0.0):
        self.change('cntrl','ntr', 1)
        self.change('cntrl','restraint_wt', restraint_wt)
        self.change('cntrl','restraintmask', '@N,CA,C')

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def genBorn(self, igb=5, rgbmax=25.0):
        self.change('cntrl','igb', igb)
        self.change('cntrl','ntb', 0)
        self.change('cntrl','ntp', 0)
        self.change('cntrl','rgbmax', rgbmax)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def time(self, time=1000.0, dt=-1): # time in ps
        if dt == -1:
            if self.cntrl_nml['ntc'] == 2 and self.cntrl_nml['ntf'] == 2:
                dt = 0.002
            else:
                dt = 0.001
        time = int(time / dt)

        self.change('cntrl','dt', dt)
        self.change('cntrl','nstlim', time)
        self.change('cntrl','imin', 0)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def heat(self, tempi=0.0, temp0=300.0, ntt=3, tautp=5.0, ig=-1, gamma_ln=5.0):
        self.constVolume()
        self.change('cntrl','tempi', tempi)
        self.change('cntrl','temp0', temp0)
        self.change('cntrl','ntt', ntt)
        self.change('cntrl','tautp', tautp)
        self.change('cntrl','ig', ig)
        self.change('cntrl','gamma_ln', gamma_ln if ntt==3 else 0)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def restart(self,ntx=5):
        self.change('cntrl','irest',1)
        self.change('cntrl','ntx',ntx)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def TI(self, clambda=0.0):
        self.change('cntrl','clambda', clambda)
        self.change('cntrl','icfe',1)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def softcore_TI(self, scalpha=0.5, scmask='', crgmask='', logdvdl=0):
        self.change('cntrl','icfe',1)
        self.change('cntrl','ifsc',1)
        self.change('cntrl','scalpha',scalpha)
        self.change('cntrl','scmask',scmask)
        self.change('cntrl','crgmask',crgmask)
        self.change('cntrl','logdvdl',logdvdl)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def minimization(self, imin=1, maxcyc=1, ncyc=10, ntmin=1):
        self.change('cntrl','imin', imin)
        self.change('cntrl','maxcyc', maxcyc)
        self.change('cntrl','ncyc', ncyc)
        self.change('cntrl','ntmin', ntmin)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def AddCard(self, title='Residues in card', cardString='RES 1'):
        self.cards.append('%s\n%s\nEND' % (title, cardString))

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



#!/bin/env python
"""
Module simtk.unit.constants

This is part of the OpenMM molecular simulation toolkit originating from
Simbios, the NIH National Center for Physics-Based Simulation of
Biological Structures at Stanford, funded under the NIH Roadmap for
Medical Research, grant U54 GM072970. See https://simtk.org.

Portions copyright (c) 2012 Stanford University and the Authors.
Authors: Christopher M. Bruns
Contributors: Peter Eastman

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from __future__ import print_function, division, absolute_import

__author__ = "Christopher M. Bruns"
__version__ = "0.5"

from .unit_definitions import *

#################
### CONSTANTS ###
#################

# codata 2006
AVOGADRO_CONSTANT_NA = 6.02214179e23 / mole
BOLTZMANN_CONSTANT_kB = 1.3806504e-23 * joule / kelvin
MOLAR_GAS_CONSTANT_R = AVOGADRO_CONSTANT_NA * BOLTZMANN_CONSTANT_kB

# From simtkcommon
SPEED_OF_LIGHT_C = 2.99792458e8 * meter / second
GRAVITATIONAL_CONSTANT_G = 6.6742e-11 * newton * meter**2 / kilogram**2

# run module directly for testing
if __name__=='__main__':
    # Test the examples in the docstrings
    import doctest, sys
    doctest.testmod(sys.modules[__name__])



"""
This module contains classes for reading and writing Amber NetCDF-style files,
including both restarts and trajectories. The NetCDF engine used here is pulled
from the scipy distribution, and depends *only* on numpy. It is available
through parmed.utils.netcdf.

This module contains objects relevant to Amber NetCDF files. The use() function
is responsible for selecting the API based on a default choice or user-selection
(the latter is really only helpful for development to ensure that all packages
work correctly---there is no difference from a user perspective). ALL
NetCDF-file manipulation that the parmed/amber package does should be
contained in this module.
"""
from __future__ import division, print_function, absolute_import

try:
    # netCDF4 is *much* faster to write NetCDF files, and can make a huge
    # difference when using it as an OpenMM reporter. So do what we can to use
    # the faster library when available
    import netCDF4 as nc
except ImportError:
    nc = None
import numpy as np
from parmed import __version__
from parmed.formats.registry import FileFormatType
from parmed import unit as u
from parmed.utils.netcdf import netcdf_file as NetCDFFile
from parmed.utils.six import add_metaclass
import warnings

@add_metaclass(FileFormatType)
class NetCDFRestart(object):
    """ Class to read or write NetCDF restart files """

    @staticmethod
    def id_format(filename):
        """ Identifies the file type as an Amber NetCDF restart file

        Parameters
        ----------
        filename : str
            Name of the file to check format for

        Returns
        -------
        is_fmt : bool
            True if it is an Amber NetCDF restart file. False otherwise

        Notes
        -----
        Remote NetCDF files cannot be loaded
        """
        if filename.startswith('http://') or filename.startswith('https://')\
                or filename.startswith('ftp://'):
            return False
        try:
            f = NetCDFFile(filename, 'r', mmap=False)
        except (TypeError, OSError):
            return False
        try:
            try:
                if f.Conventions.decode() != 'AMBERRESTART':
                    return False
            except AttributeError:
                return False
            # Passed all our tests
            return True
        finally:
            f.close()

    def __init__(self, fname, mode='r'):
        """
        Opens a NetCDF File. The main constructor should never be called
        directly.  The alternative "open_old" and "open_new" constructors should
        be called instead for reading existing or writing new NetCDF files,
        respectively.
        """
        self.closed = False
        if mode.startswith('w') and nc is not None:
            self._ncfile = nc.Dataset(fname, mode, format='NETCDF3_64BIT')
        else:
            if mode.startswith('w'):
                warnings.warn('Could not find netCDF4 module. Falling back on '
                              'scipy implementation, which can significantly '
                              'slow down simulations if used as a reporter')
            self._ncfile = NetCDFFile(fname, mode, mmap=False)

    @classmethod
    def open_new(cls, fname, natom, box, vels, title='',
                 remd=None, temp=None, remd_dimtypes=None):
        """
        Opens a new NetCDF file and sets the attributes

        Parameters
        ----------
        fname : str
            Name of the new file to open (overwritten)
        natom : int
            The number of atoms in the system
        box : bool
            Whether unit cell information is written or not
        vels : bool
            Whether velocity information is written or not
        title : str=''
            The title to write to the NetCDF restart file
        remd : str=None
            None -- No REMD information is written
            'T[emperature]' -- target temperature (or pH) will be written
            'M[ulti-D]' -- remd_dimtypes will be written
        remd_dimtypes : iterable of int=None
            Array of exchange types for each group. The length will be the REMD
            dimension (if `remd` above is "M[ulti-D]")

        Notes
        -----
        `remd` is case-insensitive, and done based on first-letter matching
        """
        if remd is not None:
            if remd[0] in 'tT':
                remd_type = 'TEMPERATURE'
                if temp is None:
                    raise ValueError('temp must be specified for '
                                     'T-REMD restarts.')
            elif remd[0] in 'mM':
                remd_type = 'MULTI'
                if remd_dimtypes is None:
                    raise ValueError('remd_dimtypes must be given for multi-D '
                                     'REMD, and must have the same length.')
                for dt in remd_dimtypes:
                    if dt not in (1, 3):
                        raise ValueError(
                                'remd_dimtypes only supports dimension types '
                                '1 and 3 currently'
                        )
                remd_dimension = len(remd_dimtypes)
            else:
                raise ValueError('remd must be None, T[emperature] or M[ulti]')
        else:
            remd_type = None
        inst = cls(fname, 'w')
        ncfile = inst._ncfile
        inst.hasbox = bool(box)
        inst.hasvels = bool(vels)
        # Assign the main attributes
        ncfile.Conventions = 'AMBERRESTART'
        ncfile.ConventionVersion = "1.0"
        ncfile.title = str(title) # Cast to avoid ScientificPython segfault
        ncfile.application = "AmberTools"
        ncfile.program = "ParmEd"
        ncfile.programVersion = str(__version__)
        # Make all of the dimensions
        ncfile.createDimension('spatial', 3)
        ncfile.createDimension('atom', natom)
        inst.spatial = 3
        inst.atom = natom
        inst.title = ncfile.title
        if box:
            ncfile.createDimension('cell_spatial', 3)
            ncfile.createDimension('label', 5)
            ncfile.createDimension('cell_angular', 3)
            inst.cell_spatial = 3
            inst.label = 5
            inst.cell_angular = 3
        if remd_type == 'MULTI':
            ncfile.createDimension('remd_dimension', remd_dimension)
            inst.remd_dimension = remd_dimension
        ncfile.createDimension('time', 1)
        # Now make the variables
        v = ncfile.createVariable('time', 'd', ('time',))
        v.units = 'picosecond'
        v = ncfile.createVariable('spatial', 'c', ('spatial',))
        v[:] = np.asarray(list('xyz'))
        if inst.hasbox:
            v = ncfile.createVariable('cell_angular', 'c',
                                            ('cell_angular', 'label'))
            v[0] = np.asarray(list('alpha'))
            v[1] = np.asarray(list('beta '))
            v[2] = np.asarray(list('gamma'))
            v = ncfile.createVariable('cell_spatial', 'c',
                                            ('cell_spatial',))
            v[0], v[1], v[2] = 'a', 'b', 'c'
            v = ncfile.createVariable('cell_lengths', 'd',
                                            ('cell_spatial',))
            v.units = 'angstrom'
            v = ncfile.createVariable('cell_angles', 'd',
                                            ('cell_angular',))
            v.units = 'degree'
        v = ncfile.createVariable('coordinates', 'd',
                                        ('atom', 'spatial'))
        v.units = 'angstrom'
        if inst.hasvels:
            v = ncfile.createVariable('velocities', 'd',
                                            ('atom', 'spatial'))
            v.units = 'angstrom/picosecond'
            v.scale_factor = np.float32(20.455)
            inst.velocity_scale = 20.455
            if nc is not None:
                v.set_auto_maskandscale(False)

        if remd_type == 'TEMPERATURE':
            v = ncfile.createVariable('temp0', 'd', ('time',))
            v.units = 'kelvin'
            v[0] = temp
        elif remd_type == 'MULTI':
            ncfile.createVariable('remd_indices', 'i',
                                        ('remd_dimension',))
            v = ncfile.createVariable('remd_dimtype', 'i',
                                        ('remd_dimension',))
            v[:] = remd_dimtypes

        return inst

    @classmethod
    def open_old(cls, fname):
        """
        Opens the NetCDF file and sets the global attributes that the file sets

        Parameters
        ----------
        fname : str
            Name of the file to read
        """
        inst = cls(fname, 'r')
        ncfile = inst._ncfile
        inst.Conventions = ncfile.Conventions.decode()
        inst.ConventionVersion = ncfile.ConventionVersion.decode()
        inst.program = ncfile.program.decode()
        inst.programVersion = ncfile.programVersion.decode()
        if hasattr(ncfile, 'application'):
            inst.application = ncfile.application.decode()
        else:
            inst.application = None
        if hasattr(ncfile, 'title'):
            inst.title = ncfile.title.decode()
        else:
            inst.title = None
        # Set up the dimensions as attributes
        for dim in ncfile.dimensions:
            # Exception for ParmEd-created ncrst files
            if dim == 'time': continue
            setattr(inst, dim, ncfile.dimensions[dim])
        inst.hasvels = 'velocities' in ncfile.variables
        inst.hasbox = ('cell_lengths' in ncfile.variables and
                       'cell_angles' in ncfile.variables)
        if inst.hasvels:
            vels = ncfile.variables['velocities']
            inst.velocity_scale = vels.scale_factor
        return inst

    parse = open_old

    @property
    def coordinates(self):
        coords = self._ncfile.variables['coordinates'][:]
        return coords.reshape((1, self.atom, 3))

    @coordinates.setter
    def coordinates(self, stuff):
        stuff = np.array(stuff, copy=False).reshape((self.atom, 3))
        self._ncfile.variables['coordinates'][:] = stuff
        self.flush()

    @property
    def velocities(self):
        if 'velocities' in self._ncfile.variables:
            vels = self._ncfile.variables['velocities'][:]
            return (vels.reshape((-1, self.atom, 3)) * self.velocity_scale)

    @velocities.setter
    def velocities(self, stuff):
        self._ncfile.variables['velocities'][:] = \
                    np.reshape(stuff, (self.atom, 3)) / self.velocity_scale
        self.flush()

    @property
    def cell_lengths(self):
        if 'cell_lengths' in self._ncfile.variables:
            return self._ncfile.variables['cell_lengths'][:]

    @cell_lengths.setter
    def cell_lengths(self, stuff):
        self._ncfile.variables['cell_lengths'][:] = np.asarray(stuff)
        self.flush()

    @property
    def cell_angles(self):
        if 'cell_angles' in self._ncfile.variables:
            return self._ncfile.variables['cell_angles'][:]

    @cell_angles.setter
    def cell_angles(self, stuff):
        self._ncfile.variables['cell_angles'][:] = np.asarray(stuff)
        self.flush()

    @property
    def box(self):
        if self.cell_lengths is not None and self.cell_angles is not None:
            leng, ang = self.cell_lengths, self.cell_angles
            return np.concatenate((leng, ang))

    @box.setter
    def box(self, stuff):
        self.cell_lengths = stuff[:3]
        self.cell_angles = stuff[3:]

    @property
    def time(self):
        return self._ncfile.variables['time'].getValue()

    @time.setter
    def time(self, stuff):
        self._ncfile.variables['time'][0] = float(stuff)
        self.flush()

    @property
    def temp0(self):
        return self._ncfile.variables['temp0'].getValue()

    @property
    def remd_indices(self):
        return self._ncfile.variables['remd_indices'][:]

    @remd_indices.setter
    def remd_indices(self, stuff):
        self._ncfile.variables['remd_indices'][:] = np.asarray(stuff, dtype='i')
        self.flush()

    @property
    def remd_dimtype(self):
        return self._ncfile.variables['remd_dimtype'][:]

    def close(self):
        if not self.closed:
            self.closed = True
            self._ncfile.close()

    def __del__(self):
        self.closed or (hasattr(self, '_ncfile') and self._ncfile.close())

    def flush(self):
        try:
            self._ncfile.flush()
        except AttributeError:
            # netCDF4.Dataset's flush method is called sync :-P
            self._ncfile.sync()

@add_metaclass(FileFormatType)
class NetCDFTraj(object):
    """ Class to read or write NetCDF restart files

    Parameters
    ----------
    fname : str
        Name of the file to open
    mode : str
        Mode to open in:
            - 'w' means write-mode
            - 'r' means read-mode

    Notes
    -----
    You should use the open_new and open_old alternative constructors instead of
    the default constructor
    """

    @staticmethod
    def id_format(filename):
        """ Identifies the file type as an Amber NetCDF trajectory file

        Parameters
        ----------
        filename : str
            Name of the file to check format for

        Returns
        -------
        is_fmt : bool
            True if it is an Amber NetCDF trajectory file. False otherwise

        Notes
        -----
        Remote NetCDF files cannot be loaded
        """
        if filename.startswith('http://') or filename.startswith('https://')\
                or filename.startswith('ftp://'):
            return False
        try:
            f = NetCDFFile(filename, 'r', mmap=False)
        except (TypeError, OSError):
            return False
        try:
            try:
                if f.Conventions.decode() != 'AMBER':
                    return False
            except AttributeError:
                return False
            # Passed all our tests
            return True
        finally:
            f.close()

    def __init__(self, fname, mode='r'):
        """ Opens a NetCDF File """
        self.closed = False
        if mode.startswith('w') and nc is not None:
            self._ncfile = nc.Dataset(fname, mode, format='NETCDF3_64BIT')
        else:
            if mode.startswith('w'):
                warnings.warn('Could not find netCDF4 module. Falling back on '
                              'scipy implementation, which can significantly '
                              'slow down simulations if used as a reporter')
            self._ncfile = NetCDFFile(fname, mode, mmap=False)

    @classmethod
    def open_new(cls, fname, natom, box, crds=True, vels=False, frcs=False,
                 remd=None, remd_dimension=None, title=''):
        """
        Opens a new NetCDF file and sets the attributes

        Parameters
        ----------
        fname : str
            Name of the new file to open (overwritten)
        natom : int
            Number of atoms in the restart
        box : bool
            Indicates if cell lengths and angles are written to the NetCDF file
        crds : bool=True
            Indicates if coordinates are written to the NetCDF file
        vels : bool=False
            Indicates if velocities are written to the NetCDF file
        frcs : bool=False
            Indicates if forces are written to the NetCDF file
        remd : str=None
            'T[emperature]' if replica temperature is written
            'M[ulti]' if Multi-D REMD information is written
            None if no REMD information is written
        remd_dimension : int=None
            If remd above is 'M[ulti]', this is how many REMD dimensions exist
        title : str=''
            The title of the NetCDF trajectory file
        """
        inst = cls(fname, 'w')
        ncfile = inst._ncfile
        if remd is not None:
            if remd[0] in 'Tt':
                inst.remd = 'TEMPERATURE'
            elif remd[0] in 'Mm':
                inst.remd = 'MULTI'
                if remd_dimension is None:
                    raise ValueError('remd_dimension must be given '
                                     'for multi-D REMD')
                inst.remd_dimension = int(remd_dimension)
            else:
                raise ValueError('remd must be T[emperature] or M[ultiD]')
        else:
            inst.remd = None
        inst.hasbox = bool(box)
        inst.hasvels = bool(vels)
        inst.hascrds = bool(crds)
        inst.hasfrcs = bool(frcs)
        # Assign the main attributes
        ncfile.Conventions = "AMBER"
        ncfile.ConventionVersion = "1.0"
        ncfile.application = "AmberTools"
        ncfile.program = "ParmEd"
        ncfile.programVersion = __version__
        ncfile.title = "ParmEd-created trajectory"
        inst.Conventions = "AMBER"
        inst.ConventionVersion = "1.0"
        inst.application = "AmberTools"
        inst.program = "ParmEd"
        inst.programVersion = __version__
        inst.title = ncfile.title
        # Create the dimensions
        ncfile.createDimension('frame', None)
        ncfile.createDimension('spatial', 3)
        ncfile.createDimension('atom', natom)
        if inst.remd == 'MULTI':
            ncfile.createDimension('remd_dimension', inst.remd_dimension)
        inst.frame, inst.spatial, inst.atom = None, 3, natom
        if inst.hasbox:
            ncfile.createDimension('cell_spatial', 3)
            ncfile.createDimension('cell_angular', 3)
            ncfile.createDimension('label', 5)
            inst.cell_spatial, inst.cell_angular, inst.label = 3, 3, 5
        # Create the variables and assign units and scaling factors
        v = ncfile.createVariable('spatial', 'c', ('spatial',))
        v[:] = np.asarray(list('xyz'))
        if inst.hasbox:
            v = ncfile.createVariable('cell_spatial', 'c',
                                            ('cell_spatial',))
            v[:] = np.asarray(list('abc'))
            v = ncfile.createVariable('cell_angular', 'c',
                                            ('cell_angular', 'label',))
            v[:] = np.asarray([list('alpha'), list('beta '), list('gamma')])
        v = ncfile.createVariable('time', 'f', ('frame',))
        v.units = 'picosecond'
        if inst.hascrds:
            v = ncfile.createVariable('coordinates', 'f',
                                            ('frame', 'atom', 'spatial'))
            v.units = 'angstrom'
            inst._last_crd_frame = 0
        if inst.hasvels:
            v = ncfile.createVariable('velocities', 'f',
                                            ('frame', 'atom', 'spatial'))
            v.units = 'angstrom/picosecond'
            inst.velocity_scale = v.scale_factor = 20.455
            inst._last_vel_frame = 0
            if nc is not None:
                v.set_auto_maskandscale(False)
        if inst.hasfrcs:
            v = ncfile.createVariable('forces', 'f',
                                            ('frame', 'atom', 'spatial'))
            v.units = 'kilocalorie/mole/angstrom'
            inst._last_frc_frame = 0
        if inst.hasbox:
            v = ncfile.createVariable('cell_lengths', 'd',
                                            ('frame', 'cell_spatial'))
            v.units = 'angstrom'
            v = ncfile.createVariable('cell_angles', 'd',
                                            ('frame', 'cell_angular'))
            v.units = 'degree'
            inst._last_box_frame = 0
        if inst.remd == 'TEMPERATURE':
            v = ncfile.createVariable('temp0', 'd', ('frame',))
            v.units = 'kelvin'
            inst._last_remd_frame = 0
        elif inst.remd == 'MULTI':
            ncfile.createVariable('remd_indices', 'i',
                                        ('frame', 'remd_dimension'))
            ncfile.createVariable('remd_dimtype', 'i',
                                        ('remd_dimension',))
            inst._last_remd_frame = 0

        inst._last_time_frame = 0

        return inst

    @classmethod
    def open_old(cls, fname):
        """
        Opens the NetCDF file and sets the global attributes that the file sets

        Parameters
        ----------
        fname : str
            File name of the trajectory to open. It must exist
        """
        inst = cls(fname, 'r')
        ncfile = inst._ncfile
        inst.Conventions = ncfile.Conventions.decode()
        inst.ConventionVersion = ncfile.ConventionVersion.decode()
        inst.program = ncfile.program.decode()
        inst.programVersion = ncfile.programVersion.decode()
        if hasattr(ncfile, 'application'):
            inst.application = ncfile.application.decode()
        else:
            inst.application = None
        if hasattr(ncfile, 'title'):
            inst.title = ncfile.title.decode()
        else:
            inst.title = None
        # Set up the dimensions as attributes
        for dim in ncfile.dimensions:
            setattr(inst, dim, ncfile.dimensions[dim])
        inst.hascrds = 'coordinates' in ncfile.variables
        inst.hasvels = 'velocities' in ncfile.variables
        inst.hasfrcs = 'forces' in ncfile.variables
        inst.hasbox = ('cell_lengths' in ncfile.variables and
                       'cell_angles' in ncfile.variables)
        if inst.hascrds:
            inst._coordinates = np.array(ncfile.variables['coordinates'][:])
        if inst.hasvels:
            try:
                scale = ncfile.variables['velocities'].scale_factor
            except AttributeError:
                scale = 1
            inst._velocities = np.array(ncfile.variables['velocities'][:])*scale
            inst.velocity_scale = scale
        if inst.hasfrcs:
            inst._forces = np.array(ncfile.variables['forces'][:])
        if inst.frame is None:
            if 'time' in ncfile.variables:
                inst.frame = len(ncfile.variables['time'][:])
            elif inst.hascrds:
                inst.frame = inst._coordinates.shape[0]
            elif inst.hasvels:
                inst.frame = inst._velocities.shape[0]
            elif inst.hasfrcs:
                inst.frame = inst._forces.shape[0]
        return inst

    @property
    def coordinates(self):
        return self._coordinates

    def add_coordinates(self, stuff):
        """
        Adds a new coordinate frame to the end of a NetCDF trajectory. This
        should only be called on objects created with the "open_new"
        constructor.

        Parameters
        ----------
        stuff : iterable of floats or distance Quantity
            This array of floats is converted into a numpy array of shape
            (natom, 3). It can be passed either in the 2-D format of
            [ [x1, y1, z1], [x2, y2, z2], ... ] or in the 1-D format of
            [x1, y1, z1, x2, y2, z2, ... ].
        """
        if u.is_quantity(stuff):
            stuff = stuff.value_in_unit(u.angstroms)
        stuff = np.asarray(stuff, dtype='f')
        self._ncfile.variables['coordinates'][self._last_crd_frame] = \
                np.reshape(stuff, (self.atom, 3))
        self._last_crd_frame += 1
        self.flush()

    @property
    def velocities(self):
        return self._velocities

    def add_velocities(self, stuff):
        """
        Adds a new velocities frame to the end of a NetCDF trajectory. This
        should only be called on objects created with the "open_new"
        constructor.

        Parameters
        ----------
        stuff : iterable of floats or distance/time Quantity
            This array of floats is converted into a numpy array of shape
            (natom, 3). It can be passed either in the 2-D format of
            [ [x1, y1, z1], [x2, y2, z2], ... ] or in the 1-D format of
            [x1, y1, z1, x2, y2, z2, ... ].
        """
        if u.is_quantity(stuff):
            stuff = stuff.value_in_unit(u.angstrom/u.picosecond)
        stuff = np.asarray(stuff)
        self._ncfile.variables['velocities'][self._last_vel_frame] = \
                np.reshape(stuff, (self.atom, 3)) / self.velocity_scale
        self._last_vel_frame += 1
        self.flush()

    @property
    def forces(self):
        return self._forces

    def add_forces(self, stuff):
        """
        Adds a new coordinate frame to the end of a NetCDF trajectory. This
        should only be called on objects created with the "open_new"
        constructor.

        Parameters
        ----------
        stuff : iterable of floats or energy/distance Quantity
            This array of floats is converted into a numpy array of shape
            (natom, 3). It can be passed either in the 2-D format of
            [ [x1, y1, z1], [x2, y2, z2], ... ] or in the 1-D format of
            [x1, y1, z1, x2, y2, z2, ... ].
        """
        if u.is_quantity(stuff):
            stuff.value_in_unit(u.kilocalories_per_mole/u.angstroms)
        self._ncfile.variables['forces'][self._last_frc_frame] = \
                np.reshape(stuff, (self.atom, 3))
        self._last_frc_frame += 1
        self.flush()

    @property
    def cell_lengths_angles(self):
        try:
            return np.hstack((self._ncfile.variables['cell_lengths'][:],
                              self._ncfile.variables['cell_angles'][:]))
        except KeyError:
            return None

    box = cell_lengths_angles

    def add_cell_lengths_angles(self, lengths, angles=None):
        """
        Adds a new cell length and angle frame to the end of a NetCDF
        trajectory.  This should only be called on objects created with the
        "open_new" constructor.

        Parameters
        ----------
        lengths : array of 3 (or 6) floats (or Quantities)
            This should be a 1-D array of 3 or 6 elements.  If 6 elements,
            `angles` should be None (below) and the first 3 elements are the box
            lengths (angstroms) and the last 3 are the box angles (degrees).
        angles : 3-item iterable = None
            These are the box angles (if lengths contains only 3 elements) in
            degrees. Must be a 1-D array of 3 elements or None if lengths
            includes angles as well.
        """
        def strip_units(x, desired_units):
            if u.is_quantity(x): return x.value_in_unit(desired_units)
            return x
        if len(lengths) == 3 and angles is None:
            raise ValueError('Both lengths and angles are required.')
        if len(lengths) == 6 and angles is not None:
            raise ValueError('Angles can be provided only once.')
        if len(lengths) != 6 and (len(lengths) != 3 or len(angles) != 3):
            raise ValueError('6 numbers expected -- 3 lengths and 3 angles.')
        if angles is None:
            angles = [strip_units(x, u.degrees) for x in lengths[3:]]
        lengths = [strip_units(x, u.angstroms) for x in lengths[:3]]
        self._ncfile.variables['cell_lengths'][self._last_box_frame] = \
                np.asarray(lengths)
        self._ncfile.variables['cell_angles'][self._last_box_frame] = \
                np.asarray(angles)
        self._last_box_frame += 1
        self.flush()

    add_box = add_cell_lengths_angles

    @property
    def time(self):
        return self._ncfile.variables['time'][:]

    def add_time(self, stuff):
        """ Adds the time to the current frame of the NetCDF file

        Parameters
        ----------
        stuff : float or time-dimension Quantity
            The time to add to the current frame
        """
        if u.is_quantity(stuff): stuff = stuff.value_in_unit(u.picoseconds)
        self._ncfile.variables['time'][self._last_time_frame] = float(stuff)
        self._last_time_frame += 1
        self.flush()

    @property
    def remd_indices(self):
        return self._ncfile.variables['remd_indices'][:]

    def add_remd_indices(self, stuff):
        """ Add REMD indices to the current frame of the NetCDF file

        Parameters
        ----------
        stuff : iterable of int
            The indices in each REMD dimension
        """
        self._ncfile.variables['remd_indices'][self._last_remd_frame] = \
                np.asarray(stuff, dtype='i')
        self._last_remd_frame += 1
        self.flush()

    @property
    def temp0(self):
        return self._ncfile.variables['temp0'][:]

    def add_temp0(self, stuff):
        """ The temperature to add to the current frame of the NetCDF file

        Parameters
        ----------
        stuff : float or temperature Quantity
            The temperature to add to the current NetCDF file
        """
        if u.is_quantity(stuff): stuff = stuff.value_in_unit(u.kelvin)
        self._ncfile.variables['temp0'][self._last_remd_frame] = float(stuff)
        self._last_remd_frame += 1
        self.flush()

    @property
    def remd_dimtype(self):
        return self._ncfile.variables['remd_dimtype'][:]

    @remd_dimtype.setter
    def remd_dimtype(self, stuff):
        self._ncfile.variables['remd_dimtype'][:] = np.asarray(stuff, dtype='i')
        self.flush()

    def close(self):
        """ Closes the NetCDF file """
        if not self.closed:
            self._ncfile.close()
            self.closed = True

    def __del__(self):
        self.closed or (hasattr(self, '_ncfile') and self._ncfile.close())

    def flush(self):
        if nc is None:
            # netCDF4.Dataset does not have a flush method
            self._ncfile.flush()



""" Tools for getting/giving file names easily for opening and saving """

import parmed.utils.six.moves.tkinter_tkfiledialog as tkFileDialog

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def file_chooser(ftype, extensions=None, set_var=None):
    """ Opens a file dialog to choose a topology file """
    if extensions is None:
        extensions = [('Amber Prmtop', '*.prmtop'), ('Amber Prmtop', '*.parm7'),
                      ('Amber Prmtop', '*.top'), ('All Files', '*')]
    fname = tkFileDialog.askopenfilename(filetypes=extensions,
                            title='Select an Amber %s file to load' % ftype)
    if set_var is not None:
        set_var.set(fname)
    return fname

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def save_file_chooser(ftype, ext, ext_opts=None):
    """ Opens a file dialog to choose a file name to save a new file as """
    if ext_opts is None:
        ext_opts = [('All files', '*')]
    return tkFileDialog.asksaveasfilename(filetypes=ext_opts,
                        defaultextension=ext, title='Save %s file as' % ftype)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~



"""
This module contains a chamber prmtop class that will read in all
parameters and allow users to manipulate that data and write a new
prmtop object.

Copyright (C) 2010 - 2015  Jason Swails

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
"""
from __future__ import absolute_import, division, print_function

import copy as _copy
import warnings
from math import pi, sqrt

from ..constants import DEG_TO_RAD, IFBOX, NATOM, NATYP, NTYPES, RAD_TO_DEG, SMALL, TINY
from ..exceptions import AmberError, AmberWarning
from ..topologyobjects import BondType, ExtraPoint, Improper, ImproperType, UreyBradley
from ..utils.six.moves import range, zip
from ._amberparm import AmberParm


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class ChamberParm(AmberParm):
    """Chamber Topology (parm7 format) class.

    Gives low, and some high, level access to topology data or
    interact with some of the high-level classes comprising the system
    topology and parameters.  The ChamberParm class uses the same
    attributes that the AmberParm class uses, and only the ones unique
    to ChamberParm will be shown below.

    Parameters
    ----------
    prm_name : str, optional
        If provided, this file is parsed and the data structures will be loaded
        from the data in this file
    xyz : str or array, optional
        If provided, the coordinates and unit cell dimensions from the provided
        Amber inpcrd/restart file will be loaded into the molecule, or the
        coordinates will be loaded from the coordinate array
    box : array, optional
        If provided, the unit cell information will be set from the provided
        unit cell dimensions (a, b, c, alpha, beta, and gamma, respectively)

    Attributes
    ----------
    LJ_14_radius : list(float)
        The same as LJ_radius, except specific for 1-4 nonbonded parameters,
        which may differ in the CHARMM force field
    LJ_14_depth : list(float)
        The same as LJ_depth, except specific for 1-4 nonbonded parameters,
        which may differ in the CHARMM force field
    urey_bradleys : TrackedList(UreyBradley)
        List of Urey-Bradley terms between two atoms in a valence angle
    impropers : TrackedList(Improper)
        List of CHARMM-style improper torsions
    cmaps : TrackedList(Cmap)
        List of coupled-torsion correction map parameters
    urey_bradley_types : TrackedList(UreyBradleyType)
        List of parameters defining the Urey-Bradley terms
    improper_types : TrackedList(Improper)
        List of parameters defining the Improper terms
    cmap_types : TrackedList(CmapType)
        List of parameters defining the CMAP terms
    chamber : bool=True
        On ChamberParm instances, this is always True to indicate that it is a
        CHAMBER-style topology file
    amoeba : bool=False
        On ChamberParm instances, this is always False to indicate that it is
        not an AMOEBA-style topology file
    has_cmap : bool
        True if CMAP parameters are present in this system; False otherwise

    See Also
    --------
    :class:`AmberParm`

    """

    _cmap_prefix = "CHARMM_"

    #===================================================

    def initialize_topology(self, xyz=None, box=None):
        """
        Initializes topology data structures, like the list of atoms, bonds,
        etc., after the topology file has been read. The following methods are
        called:
        """
        self.LJ_14_radius = []
        self.LJ_14_depth = []
        AmberParm.initialize_topology(self, xyz, box)

    #===================================================

    def _copy_lj_data(self, other):
        """ Copies Lennard-Jones lists and dicts from myself to a copy """
        super(ChamberParm, self)._copy_lj_data(other)
        other.LJ_14_radius = _copy.copy(self.LJ_14_radius)
        other.LJ_14_depth = _copy.copy(self.LJ_14_depth)
        for atom in other.atoms:
            other.LJ_14_radius[atom.nb_idx-1] = atom.atom_type.rmin_14
            other.LJ_14_depth[atom.nb_idx-1] = atom.atom_type.epsilon_14

    #===================================================

    def load_pointers(self):
        """
        Loads the data in POINTERS section into a pointers dictionary with each
        key being the pointer name according to http://ambermd.org/formats.html
        """
        AmberParm.load_pointers(self)
        # Other pointers
        nub, nubtypes = self.parm_data['CHARMM_UREY_BRADLEY_COUNT'][:2]
        self.pointers['NUB'] = nub
        self.pointers['NUBTYPES'] = nubtypes
        self.pointers['NIMPHI'] = self.parm_data['CHARMM_NUM_IMPROPERS'][0]
        self.pointers['NIMPRTYPES'] = self.parm_data['CHARMM_NUM_IMPR_TYPES'][0]

    #===================================================

    def load_structure(self):
        """
        Loads all of the topology instance variables. This is necessary if we
        actually want to modify the topological layout of our system
        (like deleting atoms)
        """
        super(ChamberParm, self).load_structure()
        self._load_urey_brad_info()
        self._load_improper_info()
        # All of our angles are urey-bradley types
        for angle in self.angles: angle.funct = 5
        super(ChamberParm, self).unchange()

    #===================================================

    @classmethod
    def from_structure(cls, struct, copy=False):
        """
        Take a Structure instance and initialize a ChamberParm instance from
        that data.

        Parameters
        ----------
        struct : Structure
            The input structure from which to construct a ChamberParm instance
        copy : bool
            If True, the input struct is deep-copied to make sure it does not
            share any objects with the original ``struct``. Default is False

        Returns
        -------
        inst : :class:`ChamberParm`
            The ChamberParm instance derived from the input structure

        Notes
        -----
        Due to the nature of the prmtop file, struct almost *always* returns a
        deep copy. The one exception is when struct is already of type
        :class:`ChamberParm`, in which case the original object is returned
        unless ``copy`` is ``True``.
        """
        if isinstance(struct, cls):
            if copy:
                return _copy.copy(struct)
            return struct
        if (struct.rb_torsions or struct.trigonal_angles or struct.pi_torsions
                or struct.out_of_plane_bends or struct.stretch_bends
                or struct.torsion_torsions or struct.multipole_frames):
            raise TypeError('ChamberParm does not support all potential terms '
                             'defined in the input Structure')
        inst = struct.copy(cls, split_dihedrals=True)
        inst.update_dihedral_exclusions()
        inst._add_missing_13_14(ignore_inconsistent_vdw=True)
        inst.pointers = {}
        inst.LJ_types = {}
        nbfixes = inst.atoms.assign_nbidx_from_types()
        # Give virtual sites a name that Amber understands
        for atom in inst.atoms:
            if isinstance(atom, ExtraPoint): atom.type = 'EP'
        # Fill the Lennard-Jones arrays/dicts
        ntyp = 0
        for atom in inst.atoms:
            inst.LJ_types[atom.type] = atom.nb_idx
            ntyp = max(ntyp, atom.nb_idx)
        inst.LJ_radius = [0 for i in range(ntyp)]
        inst.LJ_depth = [0 for i in range(ntyp)]
        inst.LJ_14_radius = [0 for i in range(ntyp)]
        inst.LJ_14_depth = [0 for i in range(ntyp)]
        for atom in inst.atoms:
            inst.LJ_radius[atom.nb_idx-1] = atom.rmin
            inst.LJ_depth[atom.nb_idx-1] = atom.epsilon
            inst.LJ_14_radius[atom.nb_idx-1] = atom.rmin_14
            inst.LJ_14_depth[atom.nb_idx-1] = atom.epsilon_14
        inst._add_standard_flags()
        inst.pointers['NATOM'] = len(inst.atoms)
        inst.parm_data['POINTERS'][NATOM] = len(inst.atoms)
        inst.box = _copy.copy(struct.box)
        if struct.box is None:
            inst.parm_data['POINTERS'][IFBOX] = 0
            inst.pointers['IFBOX'] = 0
        elif (abs(struct.box[3] - 90) > TINY or abs(struct.box[4] - 90) > TINY
                or abs(struct.box[5] - 90) > TINY):
            inst.parm_data['POINTERS'][IFBOX] = 2
            inst.pointers['IFBOX'] = 2
            inst.parm_data['BOX_DIMENSIONS'] = [struct.box[3]] + list(struct.box[:3])
        else:
            inst.parm_data['POINTERS'][IFBOX] = 1
            inst.pointers['IFBOX'] = 1
            inst.parm_data['BOX_DIMENSIONS'] = [90] + list(struct.box[:3])
        # pmemd likes to skip torsions with periodicities of 0, which may be
        # present as a way to hack entries into the 1-4 pairlist. See
        # https://github.com/ParmEd/ParmEd/pull/145 for discussion. The solution
        # here is to simply set that periodicity to 1.
        for dt in inst.dihedral_types:
            if dt.phi_k == 0 and dt.per == 0:
                dt.per = 1.0
            elif dt.per == 0:
                warnings.warn('Periodicity of 0 detected with non-zero force constant. Changing '
                              'periodicity to 1 and force constant to 0 to ensure 1-4 nonbonded '
                              'pairs are properly identified. This might cause a shift in the '
                              'energy, but will leave forces unaffected', AmberWarning)
                dt.phi_k = 0.0
                dt.per = 1.0
        inst.remake_parm()
        inst._set_nonbonded_tables(nbfixes)
        del inst.adjusts[:]
        inst.parm_data['FORCE_FIELD_TYPE'] = fftype = []
        fftype.extend([1, 'CHARMM force field: No FF information parsed...'])

        return inst

    #===================================================

    def remake_parm(self):
        """
        Fills :attr:`parm_data` from the data in the parameter and topology
        arrays (e.g., :attr:`atoms`, :attr:`bonds`, :attr:`bond_types`, ...)
        """
        # Get rid of terms containing deleted atoms and empty residues
        self.prune_empty_terms()
        self.residues.prune()
        self.rediscover_molecules()

        # Transfer information from the topology lists
        self._xfer_atom_info()
        self._xfer_residue_info()
        self._xfer_bond_info()
        self._xfer_angle_info()
        self._xfer_dihedral_info()
        self._xfer_urey_bradley_properties()
        self._xfer_improper_properties()
        self._xfer_cmap_properties()
        # Load the pointers dict
        self.load_pointers()
        # Mark atom list as unchanged
        super(ChamberParm, self).unchange()

    #===================================================

    def fill_LJ(self):
        """
        Fills the LJ_radius, LJ_depth arrays and LJ_types dictionary with data
        from LENNARD_JONES_ACOEF and LENNARD_JONES_BCOEF sections of the prmtop
        files, by undoing the canonical combining rules.
        """
        AmberParm.fill_LJ(self)

        acoef = self.parm_data['LENNARD_JONES_14_ACOEF']
        bcoef = self.parm_data['LENNARD_JONES_14_BCOEF']
        ntypes = self.pointers['NTYPES']

        self.LJ_14_radius = []  # empty LJ_radii so it can be re-filled
        self.LJ_14_depth = []   # empty LJ_depths so it can be re-filled
        one_sixth = 1.0 / 6.0 # we need to raise some numbers to the 1/6th power

        for i in range(ntypes):
            lj_index = self.parm_data["NONBONDED_PARM_INDEX"][ntypes*i+i] - 1
            if acoef[lj_index] < 1.0e-6:
                self.LJ_14_radius.append(0)
                self.LJ_14_depth.append(0)
            else:
                factor = 2 * acoef[lj_index] / bcoef[lj_index]
                self.LJ_14_radius.append(pow(factor, one_sixth) * 0.5)
                self.LJ_14_depth.append(bcoef[lj_index] / 2 / factor)

    #===================================================

    def recalculate_LJ(self):
        """
        Takes the values of the LJ_radius and LJ_depth arrays and recalculates
        the LENNARD_JONES_A/BCOEF topology sections from the canonical
        combining rules.
        """
        AmberParm.recalculate_LJ(self)
        ntypes = self.pointers['NTYPES']
        acoef = self.parm_data['LENNARD_JONES_14_ACOEF']
        bcoef = self.parm_data['LENNARD_JONES_14_BCOEF']
        for i in range(ntypes):
            for j in range(i, ntypes):
                index = self.parm_data['NONBONDED_PARM_INDEX'][ntypes*i+j] - 1
                rij = self.LJ_14_radius[i] + self.LJ_14_radius[j]
                wdij = sqrt(self.LJ_14_depth[i] * self.LJ_14_depth[j])
                acoef[index] = wdij * rij ** 12
                bcoef[index] = 2 * wdij * rij**6

    #===================================================

    @property
    def chamber(self):
        return True

    @property
    def amoeba(self):
        return False

    #===========  PRIVATE INSTANCE METHODS  ============

    def _load_urey_brad_info(self):
        """ Loads the Urey-Bradley types and array """
        del self.urey_bradleys[:]
        del self.urey_bradley_types[:]
        for k, req in zip(self.parm_data['CHARMM_UREY_BRADLEY_FORCE_CONSTANT'],
                          self.parm_data['CHARMM_UREY_BRADLEY_EQUIL_VALUE']):
            self.urey_bradley_types.append(
                    BondType(k, req, self.urey_bradley_types)
            )
        it = iter(self.parm_data['CHARMM_UREY_BRADLEY'])
        for i, j, k in zip(it, it, it):
            self.urey_bradleys.append(
                    UreyBradley(self.atoms[i-1], self.atoms[j-1],
                                self.urey_bradley_types[k-1])
            )

    #===================================================

    def _load_improper_info(self):
        """ Loads the CHARMM Improper types and array """
        del self.impropers[:]
        del self.improper_types[:]
        for k, eq in zip(self.parm_data['CHARMM_IMPROPER_FORCE_CONSTANT'],
                         self.parm_data['CHARMM_IMPROPER_PHASE']):
            # Previous versions of ParmEd stored improper phases as degrees,
            # whereas it should really be stored in radians. So do a simple
            # heuristic check to see if a conversion is necessary so we support
            # all versions.
            eq = eq * RAD_TO_DEG if abs(eq) <= 2*pi else eq
            self.improper_types.append(
                    ImproperType(k, eq, self.improper_types)
            )
        it = iter(self.parm_data['CHARMM_IMPROPERS'])
        for i, j, k, l, m in zip(it, it, it, it, it):
            self.impropers.append(
                    Improper(self.atoms[i-1], self.atoms[j-1], self.atoms[k-1],
                             self.atoms[l-1], self.improper_types[m-1])
            )
        # Make sure that if we have a comment in the CHARMM impropers, we fix it
        # to say the units are in radians
        for i in range(len(self.parm_comments.get('CHARMM_IMPROPER_PHASE', []))):
            comment = self.parm_comments['CHARMM_IMPROPER_PHASE'][i]
            if 'degrees' in comment:
                self.parm_comments['CHARMM_IMPROPER_PHASE'][i] = \
                        comment.replace('degrees', 'radians')

    #===================================================

    def _check_section_lengths(self):
        """ Make sure each section has the necessary number of entries """
        super(ChamberParm, self)._check_section_lengths()

        def check_length(key, length, required=True):
            if not required and not key in self.parm_data: return
            if len(self.parm_data[key]) != length:
                raise AmberError('FLAG %s has %d elements; expected %d' %
                                 (key, len(self.parm_data[key]), length))

        check_length('CHARMM_UREY_BRADLEY_COUNT', 2)
        check_length('CHARMM_UREY_BRADLEY', self.pointers['NUB']*3)
        check_length('CHARMM_UREY_BRADLEY_FORCE_CONSTANT',
                     self.pointers['NUBTYPES'])
        check_length('CHARMM_UREY_BRADLEY_EQUIL_VALUE',
                     self.pointers['NUBTYPES'])
        check_length('CHARMM_NUM_IMPROPERS', 1)
        check_length('CHARMM_IMPROPERS', self.pointers['NIMPHI']*5)
        check_length('CHARMM_NUM_IMPR_TYPES', 1)
        check_length('CHARMM_IMPROPER_FORCE_CONSTANT',
                     self.pointers['NIMPRTYPES'])
        check_length('CHARMM_IMPROPER_PHASE', self.pointers['NIMPRTYPES'])

        ntypes = self.pointers['NTYPES']
        check_length('LENNARD_JONES_14_ACOEF', ntypes*(ntypes+1)//2)
        check_length('LENNARD_JONES_14_BCOEF', ntypes*(ntypes+1)//2)

    #===================================================

    def _xfer_urey_bradley_properties(self):
        """
        Sets the various topology file section data from the Urey-Bradley arrays
        """
        data = self.parm_data
        for urey_type in self.urey_bradley_types:
            urey_type.used = False
        for urey in self.urey_bradleys:
            urey.type.used = True
        self.urey_bradley_types.prune_unused()
        data['CHARMM_UREY_BRADLEY_FORCE_CONSTANT'] = \
                [type.k for type in self.urey_bradley_types]
        data['CHARMM_UREY_BRADLEY_EQUIL_VALUE'] = \
                [type.req for type in self.urey_bradley_types]
        data['CHARMM_UREY_BRADLEY'] = bond_array = []
        for urey in self.urey_bradleys:
            bond_array.extend([urey.atom1.idx+1, urey.atom2.idx+1,
                               urey.type.idx+1])
        nub = len(self.urey_bradleys)
        nubt = len(self.urey_bradley_types)
        data['CHARMM_UREY_BRADLEY_COUNT'] = [nub, nubt]
        self.pointers['NUB'] = nub
        self.pointers['NUBTYPES'] = nubt

    #===================================================

    def _xfer_improper_properties(self):
        """ Sets the topology file section data from the improper arrays """
        data = self.parm_data
        for improper_type in self.improper_types:
            improper_type.used = False
        for improper in self.impropers:
            improper.type.used = True
        self.improper_types.prune_unused()
        data['CHARMM_IMPROPER_FORCE_CONSTANT'] = \
                [type.psi_k for type in self.improper_types]
        data['CHARMM_IMPROPER_PHASE'] = \
                [type.psi_eq*DEG_TO_RAD for type in self.improper_types]
        data['CHARMM_IMPROPERS'] = improper_array = []
        for imp in self.impropers:
            improper_array.extend([imp.atom1.idx+1, imp.atom2.idx+1,
                                   imp.atom3.idx+1, imp.atom4.idx+1,
                                   imp.type.idx+1])
        data['CHARMM_NUM_IMPROPERS'] = [len(self.impropers)]
        data['CHARMM_NUM_IMPR_TYPES'] = [len(self.improper_types)]
        self.pointers['NIMPHI'] = len(improper_array)
        self.pointers['NIMPRTYPES'] = len(self.improper_types)

    #===================================================

    def _add_standard_flags(self):
        """ Adds all of the standard flags to the parm_data array """
        self.set_version()
        self.add_flag('CTITLE', '20a4', num_items=0)
        self.add_flag('POINTERS', '10I8', num_items=31)
        self.add_flag('FORCE_FIELD_TYPE', 'i2,a78', num_items=0)
        self.add_flag('ATOM_NAME', '20a4', num_items=0)
        self.add_flag('CHARGE', '3E24.16', num_items=0,
                      comments=['Atomic charge multiplied by sqrt(332.0716D0) (CCELEC)'])
        self.add_flag('ATOMIC_NUMBER', '10I8', num_items=0)
        self.add_flag('MASS', '5E16.8', num_items=0)
        self.add_flag('ATOM_TYPE_INDEX', '10I8', num_items=0)
        self.add_flag('NUMBER_EXCLUDED_ATOMS', '10I8', num_items=0)
        self.add_flag('NONBONDED_PARM_INDEX', '10I8', num_items=0)
        self.add_flag('RESIDUE_LABEL', '20a4', num_items=0)
        self.add_flag('RESIDUE_POINTER', '10I8', num_items=0)
        self.add_flag('BOND_FORCE_CONSTANT', '5E16.8', num_items=0)
        self.add_flag('BOND_EQUIL_VALUE', '5E16.8', num_items=0)
        self.add_flag('ANGLE_FORCE_CONSTANT', '5E16.8', num_items=0)
        self.add_flag('ANGLE_EQUIL_VALUE', '3E25.17', num_items=0)
        self.add_flag('CHARMM_UREY_BRADLEY_COUNT', '2I8', num_items=2,
                      comments=['V(ub) = K_ub(r_ik - R_ub)**2',
                                'Number of Urey Bradley terms and types'])
        self.add_flag('CHARMM_UREY_BRADLEY', '10I8', num_items=0,
                      comments=['List of the two atoms and its parameter index',
                                'in each UB term: i,k,index'])
        self.add_flag('CHARMM_UREY_BRADLEY_FORCE_CONSTANT', '5E16.8', num_items=0,
                      comments=['K_ub: kcal/mol/A**2'])
        self.add_flag('CHARMM_UREY_BRADLEY_EQUIL_VALUE', '5E16.8', num_items=0,
                      comments=['r_ub: A'])
        self.add_flag('DIHEDRAL_FORCE_CONSTANT', '5E16.8', num_items=0)
        self.add_flag('DIHEDRAL_PERIODICITY', '5E16.8', num_items=0)
        self.add_flag('DIHEDRAL_PHASE', '5E16.8', num_items=0)
        self.add_flag('SCEE_SCALE_FACTOR', '5E16.8', num_items=0)
        self.add_flag('SCNB_SCALE_FACTOR', '5E16.8', num_items=0)
        self.add_flag('CHARMM_NUM_IMPROPERS', '10I8', num_items=0,
                      comments=['Number of terms contributing to the',
                                'quadratic four atom improper energy term:',
                                'V(improper) = K_psi(psi - psi_0)**2'])
        self.add_flag('CHARMM_IMPROPERS', '10I8', num_items=0,
                      comments=['List of the four atoms in each improper term',
                                'i,j,k,l,index  i,j,k,l,index',
                                'where index is into the following two lists:',
                                'CHARMM_IMPROPER_{FORCE_CONSTANT,IMPROPER_PHASE}'])
        self.add_flag('CHARMM_NUM_IMPR_TYPES', '1I8', num_items=1,
                      comments=['Number of unique parameters contributing to the',
                                'quadratic four atom improper energy term'])
        self.add_flag('CHARMM_IMPROPER_FORCE_CONSTANT', '5E16.8', num_items=0,
                      comments=['K_psi: kcal/mole/rad**2'])
        self.add_flag('CHARMM_IMPROPER_PHASE', '5E16.8', num_items=0, comments=['psi: radians'])
        natyp = self.pointers['NATYP'] = self.parm_data['POINTERS'][NATYP] = 1
        self.add_flag('SOLTY', '5E16.8', num_items=natyp)
        self.add_flag('LENNARD_JONES_ACOEF', '3E24.16', num_items=0)
        self.add_flag('LENNARD_JONES_BCOEF', '3E24.16', num_items=0)
        self.add_flag('LENNARD_JONES_14_ACOEF', '3E24.16', num_items=0)
        self.add_flag('LENNARD_JONES_14_BCOEF', '3E24.16', num_items=0)
        self.add_flag('BONDS_INC_HYDROGEN', '10I8', num_items=0)
        self.add_flag('BONDS_WITHOUT_HYDROGEN', '10I8', num_items=0)
        self.add_flag('ANGLES_INC_HYDROGEN', '10I8', num_items=0)
        self.add_flag('ANGLES_WITHOUT_HYDROGEN', '10I8', num_items=0)
        self.add_flag('DIHEDRALS_INC_HYDROGEN', '10I8', num_items=0)
        self.add_flag('DIHEDRALS_WITHOUT_HYDROGEN', '10I8', num_items=0)
        self.add_flag('EXCLUDED_ATOMS_LIST', '10I8', num_items=0)
        self.add_flag('HBOND_ACOEF', '5E16.8', num_items=0)
        self.add_flag('HBOND_BCOEF', '5E16.8', num_items=0)
        self.add_flag('HBCUT', '5E16.8', num_items=0)
        self.add_flag('AMBER_ATOM_TYPE', '20a4', num_items=0)
        self.add_flag('TREE_CHAIN_CLASSIFICATION', '20a4', num_items=0)
        self.add_flag('JOIN_ARRAY', '10I8', num_items=0)
        self.add_flag('IROTAT', '10I8', num_items=0)
        if self.has_cmap:
            self.add_flag(self._cmap_prefix + 'CMAP_COUNT', '2I8', num_items=2,
                          comments=['Number of CMAP terms, number of unique CMAP parameters'])
            self.add_flag(self._cmap_prefix + 'CMAP_RESOLUTION', '20I4', num_items=0,
                          comments=['Number of steps along each phi/psi CMAP axis',
                                    'for each CMAP_PARAMETER grid'])
            self.add_flag(self._cmap_prefix + 'CMAP_INDEX', '6I8', num_items=0,
                          comments=['Atom index i,j,k,l,m of the cross term',
                                    'and then pointer to CMAP_PARAMETER_n'])
        if self.box is not None:
            self.add_flag('SOLVENT_POINTERS', '3I8', num_items=3)
            self.add_flag('ATOMS_PER_MOLECULE', '10I8', num_items=0)
            self.add_flag('BOX_DIMENSIONS', '5E16.8', num_items=4)
        self.add_flag('RADIUS_SET', '1a80', num_items=1)
        self.add_flag('RADII', '5E16.8', num_items=0)
        self.add_flag('SCREEN', '5E16.8', num_items=0)
        self.add_flag('IPOL', '1I8', num_items=1)

    #===================================================

    def _set_nonbonded_tables(self, nbfixes=None):
        """ Sets the tables of Lennard-Jones nonbonded interaction pairs """
        from parmed.tools.actions import addLJType
        data = self.parm_data
        ntypes = data['POINTERS'][NTYPES]
        ntypes2 = ntypes * ntypes
        # Set up the index lookup tables (not a unique solution)
        data['NONBONDED_PARM_INDEX'] = [0 for i in range(ntypes2)]
        holder = [0 for i in range(ntypes2)]
        idx = 0
        for i in range(ntypes):
            for j in range(i+1):
                idx += 1
                holder[ntypes*i+j] = holder[ntypes*j+i] = idx
        idx = 0
        for i in range(ntypes):
            for j in range(ntypes):
                data['NONBONDED_PARM_INDEX'][idx] = holder[ntypes*i+j]
                idx += 1
        nttyp = ntypes * (ntypes + 1) // 2
        # Now build the Lennard-Jones arrays
        data['LENNARD_JONES_14_ACOEF'] = [0 for i in range(nttyp)]
        data['LENNARD_JONES_14_BCOEF'] = [0 for i in range(nttyp)]
        data['LENNARD_JONES_ACOEF'] = [0 for i in range(nttyp)]
        data['LENNARD_JONES_BCOEF'] = [0 for i in range(nttyp)]
        self.recalculate_LJ()
        # Now make any NBFIX modifications we had
        if nbfixes is not None:
            for i, fix in enumerate(nbfixes):
                for terms in fix:
                    j, rmin, eps, rmin14, eps14 = terms
                    i, j = min(i, j-1), max(i, j-1)
                    eps = abs(eps)
                    eps14 = abs(eps14)
                    idx = data['NONBONDED_PARM_INDEX'][ntypes*i+j] - 1
                    data['LENNARD_JONES_ACOEF'][idx] = eps * rmin**12
                    data['LENNARD_JONES_BCOEF'][idx] = 2 * eps * rmin**6
                    data['LENNARD_JONES_14_ACOEF'][idx] = eps14 * rmin14**12
                    data['LENNARD_JONES_14_BCOEF'][idx] = 2 * eps14 * rmin14**6
        # If we had an explicit set of exceptions, we need to implement all of
        # those exclusions. The electrostatic component of that exception is
        # already handled (since a scaling factor *can* represent the full
        # flexibility of electrostatic exceptions). The vdW component must be
        # handled as 1-4 A- and B-coefficients. If vdW types are too compressed
        # for this to be handled correctly, use "addLJType" to expand the types
        # by 1 (so the tables only get as big as they *need* to get, to make
        # sure they continue to fit in CUDA shared memory).
        #
        # The way we do this is to fill all of the elements with None, then fill
        # them in as we walk through the 1-4 exceptions. If we hit a case where
        # the target element is not None and *doesn't* equal the computed A- and
        # B-coefficients, we have to expand our type list (assume all type
        # names will have the same L-J parameters, which has been a fair
        # assumption in my experience).
        if not self.adjusts: return
        for i in range(len(data['LENNARD_JONES_14_ACOEF'])):
            data['LENNARD_JONES_14_ACOEF'][i] = None
            data['LENNARD_JONES_14_BCOEF'][i] = None
        ii = 0
        replaced_atoms = set()
        while True:
            needed_split = False
            for pair in self.adjusts:
                a1, a2 = pair.atom1, pair.atom2
                i, j = sorted([a1.nb_idx - 1, a2.nb_idx - 1])
                idx = data['NONBONDED_PARM_INDEX'][ntypes*i+j] - 1
                eps = pair.type.epsilon
                rmin = pair.type.rmin
                rmin6 = rmin * rmin * rmin * rmin * rmin * rmin
                acoef = eps * rmin6*rmin6
                bcoef = 2 * eps * rmin6
                if data['LENNARD_JONES_14_ACOEF'][idx] is not None:
                    if abs(data['LENNARD_JONES_14_ACOEF'][idx] - acoef) > SMALL:
                        # Need to split out another type
                        needed_split = True
                        assert a1 not in replaced_atoms or a2 not in replaced_atoms
                        # Only add each atom as a new type ONCE
                        if a1 in replaced_atoms:
                            mask = '@%d' % (a2.idx+1)
                            replaced_atoms.add(a2)
                        else:
                            mask = '@%d' % (a1.idx+1)
                            replaced_atoms.add(a1)
                        addLJType(self, mask, radius_14=0, epsilon_14=0).execute()
                        ntypes += 1
                        # None-out all of the added terms
                        j = ntypes - 1
                        for i in range(j):
                            idx2 = data['NONBONDED_PARM_INDEX'][ntypes*i+j] - 1
                            data['LENNARD_JONES_14_ACOEF'][idx2] = None
                            data['LENNARD_JONES_14_BCOEF'][idx2] = None
                        # We can stop here, since the next loop through the
                        # explicit exclusions will fill this in
                else:
                    data['LENNARD_JONES_14_ACOEF'][idx] = acoef
                    data['LENNARD_JONES_14_BCOEF'][idx] = bcoef
            ii += 1
            if not needed_split:
                break
            # The following should never happen
            assert ii <= len(self.atoms)+1, 'Could not resolve all exceptions. ' \
                   'Some unexpected problem with the algorithm'
        # Now go through and change all None's to 0s, as these terms won't be
        # used for any exceptions, anyway
        for i, item in enumerate(data['LENNARD_JONES_14_ACOEF']):
            if item is None:
                assert data['LENNARD_JONES_14_BCOEF'][i] is None, \
                       'A- and B- coefficients must be in lock-step!'
                data['LENNARD_JONES_14_ACOEF'][i] = 0.0
                data['LENNARD_JONES_14_BCOEF'][i] = 0.0

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def ConvertFromPSF(struct, params, title=''):
    """
    This function instantiates a ChamberParm instance from a data structure
    instantiated by a CHARMM PSF.

    Parameters
    ----------
    struct : Structure
        The structure object (typically loaded from a PSF file)
    params : CharmmParameterSet
        The parameter set describing the parameters of the input struct
    title : str=''
        The title to assign to the topology file

    Returns
    -------
    ChamberParm
        ChamberParm instance with all parameters loaded
    """
    # Make sure all atom types are strings, not integers
    int_starting = str(struct.atoms[0].atom_type) != struct.atoms[0].type
    if int_starting:
        for atom in struct.atoms:
            atom.type = str(atom.atom_type)
    parm = ChamberParm.from_structure(struct)
    parm.parm_data['FORCE_FIELD_TYPE'] = fftype = []
    if params.parametersets == []:
        params.parametersets.append('')
    for pset in params.parametersets:
        if 'CHARMM' not in pset: # needed to trigger "charmm_active"...
            pset = 'CHARMM: %s' % pset
        fftype.extend([len(params.parametersets), pset])

    # Convert atom types back to integers if that's how they started
    if int_starting:
        for atom in struct.atoms:
            atom.type = int(atom.atom_type)
    return parm
